import unittest
import os
import sys
from fontTools import t1Lib
from fontTools.pens.basePen import NullPen
from fontTools.misc.psCharStrings import T1CharString
import random


CWD = os.path.abspath(os.path.dirname(__file__))
DATADIR = os.path.join(CWD, "data")
# I used `tx` to convert PFA to LWFN (stored in the data fork)
LWFN = os.path.join(DATADIR, "TestT1-Regular.lwfn")
PFA = os.path.join(DATADIR, "TestT1-Regular.pfa")
PFB = os.path.join(DATADIR, "TestT1-Regular.pfb")
WEIRD_ZEROS = os.path.join(DATADIR, "TestT1-weird-zeros.pfa")
# ellipsis is hinted with 55 131 296 131 537 131 vstem3 0 122 hstem
ELLIPSIS_HINTED = os.path.join(DATADIR, "TestT1-ellipsis-hinted.pfa")


class FindEncryptedChunksTest(unittest.TestCase):
    def test_findEncryptedChunks(self):
        with open(PFA, "rb") as f:
            data = f.read()
        chunks = t1Lib.findEncryptedChunks(data)
        self.assertEqual(len(chunks), 3)
        self.assertFalse(chunks[0][0])
        # the second chunk is encrypted
        self.assertTrue(chunks[1][0])
        self.assertFalse(chunks[2][0])

    def test_findEncryptedChunks_weird_zeros(self):
        with open(WEIRD_ZEROS, "rb") as f:
            data = f.read()

        # Just assert that this doesn't raise any exception for not finding the
        # end of eexec
        t1Lib.findEncryptedChunks(data)


class DecryptType1Test(unittest.TestCase):
    def test_decryptType1(self):
        with open(PFA, "rb") as f:
            data = f.read()
        decrypted = t1Lib.decryptType1(data)
        self.assertNotEqual(decrypted, data)


class ReadWriteTest(unittest.TestCase):
    def test_read_pfa_write_pfb(self):
        font = t1Lib.T1Font(PFA)
        data = self.write(font, "PFB")
        self.assertEqual(font.getData(), data)

    def test_read_and_parse_pfa_write_pfb(self):
        font = t1Lib.T1Font(PFA)
        font.parse()
        saved_font = self.write(font, "PFB", dohex=False, doparse=True)
        self.assertTrue(same_dicts(font.font, saved_font))

    def test_read_pfb_write_pfa(self):
        font = t1Lib.T1Font(PFB)
        # 'OTHER' == 'PFA'
        data = self.write(font, "OTHER", dohex=True)
        self.assertEqual(font.getData(), data)

    def test_read_and_parse_pfb_write_pfa(self):
        font = t1Lib.T1Font(PFB)
        font.parse()
        # 'OTHER' == 'PFA'
        saved_font = self.write(font, "OTHER", dohex=True, doparse=True)
        self.assertTrue(same_dicts(font.font, saved_font))

    def test_read_with_path(self):
        import pathlib

        font = t1Lib.T1Font(pathlib.Path(PFB))

    @staticmethod
    def write(font, outtype, dohex=False, doparse=False):
        temp = os.path.join(DATADIR, "temp." + outtype.lower())
        try:
            font.saveAs(temp, outtype, dohex=dohex)
            newfont = t1Lib.T1Font(temp)
            if doparse:
                newfont.parse()
                data = newfont.font
            else:
                data = newfont.getData()
        finally:
            if os.path.exists(temp):
                os.remove(temp)
        return data


class T1FontTest(unittest.TestCase):
    def test_parse_lwfn(self):
        # the extended attrs are lost on git so we can't auto-detect 'LWFN'
        font = t1Lib.T1Font(LWFN, kind="LWFN")
        font.parse()
        self.assertEqual(font["FontName"], "TestT1-Regular")
        self.assertTrue("Subrs" in font["Private"])

    def test_parse_pfa(self):
        font = t1Lib.T1Font(PFA)
        font.parse()
        self.assertEqual(font["FontName"], "TestT1-Regular")
        self.assertTrue("Subrs" in font["Private"])

    def test_parse_pfb(self):
        font = t1Lib.T1Font(PFB)
        font.parse()
        self.assertEqual(font["FontName"], "TestT1-Regular")
        self.assertTrue("Subrs" in font["Private"])

    def test_getGlyphSet(self):
        font = t1Lib.T1Font(PFA)
        glyphs = font.getGlyphSet()
        i = random.randrange(len(glyphs))
        aglyph = list(glyphs.values())[i]
        self.assertTrue(hasattr(aglyph, "draw"))
        self.assertFalse(hasattr(aglyph, "width"))
        aglyph.draw(NullPen())
        self.assertTrue(hasattr(aglyph, "width"))


class EditTest(unittest.TestCase):
    def test_edit_pfa(self):
        font = t1Lib.T1Font(PFA)
        ellipsis = font.getGlyphSet()["ellipsis"]
        ellipsis.decompile()
        program = []
        for v in ellipsis.program:
            try:
                program.append(int(v))
            except:
                program.append(v)
                if v == "hsbw":
                    hints = [55, 131, 296, 131, 537, 131, "vstem3", 0, 122, "hstem"]
                    program.extend(hints)
        ellipsis.program = program
        # 'OTHER' == 'PFA'
        saved_font = self.write(font, "OTHER", dohex=True, doparse=True)
        hinted_font = t1Lib.T1Font(ELLIPSIS_HINTED)
        hinted_font.parse()
        self.assertTrue(same_dicts(hinted_font.font, saved_font))

    @staticmethod
    def write(font, outtype, dohex=False, doparse=False):
        temp = os.path.join(DATADIR, "temp." + outtype.lower())
        try:
            font.saveAs(temp, outtype, dohex=dohex)
            newfont = t1Lib.T1Font(temp)
            if doparse:
                newfont.parse()
                data = newfont.font
            else:
                data = newfont.getData()
        finally:
            if os.path.exists(temp):
                os.remove(temp)
        return data


def same_dicts(dict1, dict2):
    if dict1.keys() != dict2.keys():
        return False
    for key, value in dict1.items():
        if isinstance(value, dict):
            if not same_dicts(value, dict2[key]):
                return False
        elif isinstance(value, list):
            if len(value) != len(dict2[key]):
                return False
            for elem1, elem2 in zip(value, dict2[key]):
                if isinstance(elem1, T1CharString):
                    elem1.compile()
                    elem2.compile()
                    if elem1.bytecode != elem2.bytecode:
                        return False
                else:
                    if elem1 != elem2:
                        return False
        elif isinstance(value, T1CharString):
            value.compile()
            dict2[key].compile()
            if value.bytecode != dict2[key].bytecode:
                return False
        else:
            if value != dict2[key]:
                return False
    return True


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



import os

from fontTools.misc.loggingTools import CapturingLogHandler
from fontTools.cu2qu.ufo import (
    fonts_to_quadratic,
    font_to_quadratic,
    glyphs_to_quadratic,
    glyph_to_quadratic,
    logger,
    CURVE_TYPE_LIB_KEY,
)
from fontTools.cu2qu.errors import (
    IncompatibleSegmentNumberError,
    IncompatibleSegmentTypesError,
    IncompatibleFontsError,
)

import pytest


ufoLib2 = pytest.importorskip("ufoLib2")

DATADIR = os.path.join(os.path.dirname(__file__), "data")

TEST_UFOS = [
    os.path.join(DATADIR, "RobotoSubset-Regular.ufo"),
    os.path.join(DATADIR, "RobotoSubset-Bold.ufo"),
]


@pytest.fixture
def fonts():
    return [ufoLib2.Font.open(ufo) for ufo in TEST_UFOS]


class FontsToQuadraticTest(object):
    def test_modified(self, fonts):
        # previously this method returned True/False, now it returns a set of modified
        # glyph names.
        modified = fonts_to_quadratic(fonts)
        # the first assertion continues to work whether the return value is a bool/set
        # so the change is backward compatible
        assert modified
        assert len(modified) > 0
        assert "B" in modified

    def test_stats(self, fonts):
        stats = {}
        fonts_to_quadratic(fonts, stats=stats)
        assert stats == {"1": 1, "2": 79, "3": 130, "4": 2}

    def test_dump_stats(self, fonts):
        with CapturingLogHandler(logger, "INFO") as captor:
            fonts_to_quadratic(fonts, dump_stats=True)
        assert captor.assertRegex("New spline lengths:")

    def test_remember_curve_type_quadratic(self, fonts):
        fonts_to_quadratic(fonts, remember_curve_type=True)
        assert fonts[0].lib[CURVE_TYPE_LIB_KEY] == "quadratic"
        with CapturingLogHandler(logger, "INFO") as captor:
            fonts_to_quadratic(fonts, remember_curve_type=True)
        assert captor.assertRegex("already converted")

    def test_remember_curve_type_mixed(self, fonts):
        fonts_to_quadratic(fonts, remember_curve_type=True, all_quadratic=False)
        assert fonts[0].lib[CURVE_TYPE_LIB_KEY] == "mixed"
        with CapturingLogHandler(logger, "INFO") as captor:
            fonts_to_quadratic(fonts, remember_curve_type=True)
        assert captor.assertRegex("already converted")

    def test_no_remember_curve_type(self, fonts):
        assert CURVE_TYPE_LIB_KEY not in fonts[0].lib
        fonts_to_quadratic(fonts, remember_curve_type=False)
        assert CURVE_TYPE_LIB_KEY not in fonts[0].lib

    def test_different_glyphsets(self, fonts):
        del fonts[0]["a"]
        assert "a" not in fonts[0]
        assert "a" in fonts[1]
        assert fonts_to_quadratic(fonts)

    def test_max_err_em_float(self, fonts):
        stats = {}
        fonts_to_quadratic(fonts, max_err_em=0.002, stats=stats)
        assert stats == {"1": 5, "2": 193, "3": 14}

    def test_max_err_em_list(self, fonts):
        stats = {}
        fonts_to_quadratic(fonts, max_err_em=[0.002, 0.002], stats=stats)
        assert stats == {"1": 5, "2": 193, "3": 14}

    def test_max_err_float(self, fonts):
        stats = {}
        fonts_to_quadratic(fonts, max_err=4.096, stats=stats)
        assert stats == {"1": 5, "2": 193, "3": 14}

    def test_max_err_list(self, fonts):
        stats = {}
        fonts_to_quadratic(fonts, max_err=[4.096, 4.096], stats=stats)
        assert stats == {"1": 5, "2": 193, "3": 14}

    def test_both_max_err_and_max_err_em(self, fonts):
        with pytest.raises(TypeError, match="Only one .* can be specified"):
            fonts_to_quadratic(fonts, max_err=1.000, max_err_em=0.001)

    def test_single_font(self, fonts):
        assert font_to_quadratic(fonts[0], max_err_em=0.002, reverse_direction=True)
        assert font_to_quadratic(
            fonts[1], max_err_em=0.002, reverse_direction=True, all_quadratic=False
        )


class GlyphsToQuadraticTest(object):
    @pytest.mark.parametrize(
        ["glyph", "expected"],
        [("A", False), ("a", True)],  # contains no curves, it is not modified
        ids=["lines-only", "has-curves"],
    )
    def test_modified(self, fonts, glyph, expected):
        glyphs = [f[glyph] for f in fonts]
        assert glyphs_to_quadratic(glyphs) == expected

    def test_stats(self, fonts):
        stats = {}
        glyphs_to_quadratic([f["a"] for f in fonts], stats=stats)
        assert stats == {"2": 1, "3": 7, "4": 3, "5": 1}

    def test_max_err_float(self, fonts):
        glyphs = [f["a"] for f in fonts]
        stats = {}
        glyphs_to_quadratic(glyphs, max_err=4.096, stats=stats)
        assert stats == {"2": 11, "3": 1}

    def test_max_err_list(self, fonts):
        glyphs = [f["a"] for f in fonts]
        stats = {}
        glyphs_to_quadratic(glyphs, max_err=[4.096, 4.096], stats=stats)
        assert stats == {"2": 11, "3": 1}

    def test_reverse_direction(self, fonts):
        glyphs = [f["A"] for f in fonts]
        assert glyphs_to_quadratic(glyphs, reverse_direction=True)

    def test_single_glyph(self, fonts):
        assert glyph_to_quadratic(fonts[0]["a"], max_err=4.096, reverse_direction=True)

    @pytest.mark.parametrize(
        ["outlines", "exception", "message"],
        [
            [
                [
                    [
                        ("moveTo", ((0, 0),)),
                        ("curveTo", ((1, 1), (2, 2), (3, 3))),
                        ("curveTo", ((4, 4), (5, 5), (6, 6))),
                        ("closePath", ()),
                    ],
                    [
                        ("moveTo", ((7, 7),)),
                        ("curveTo", ((8, 8), (9, 9), (10, 10))),
                        ("closePath", ()),
                    ],
                ],
                IncompatibleSegmentNumberError,
                "have different number of segments",
            ],
            [
                [
                    [
                        ("moveTo", ((0, 0),)),
                        ("curveTo", ((1, 1), (2, 2), (3, 3))),
                        ("closePath", ()),
                    ],
                    [
                        ("moveTo", ((4, 4),)),
                        ("lineTo", ((5, 5),)),
                        ("closePath", ()),
                    ],
                ],
                IncompatibleSegmentTypesError,
                "have incompatible segment types",
            ],
        ],
        ids=[
            "unequal-length",
            "different-segment-types",
        ],
    )
    def test_incompatible_glyphs(self, outlines, exception, message):
        glyphs = []
        for i, outline in enumerate(outlines):
            glyph = ufoLib2.objects.Glyph("glyph%d" % i)
            pen = glyph.getPen()
            for operator, args in outline:
                getattr(pen, operator)(*args)
            glyphs.append(glyph)
        with pytest.raises(exception) as excinfo:
            glyphs_to_quadratic(glyphs)
        assert excinfo.match(message)

    def test_incompatible_fonts(self):
        font1 = ufoLib2.Font()
        font1.info.unitsPerEm = 1000
        glyph1 = font1.newGlyph("a")
        pen1 = glyph1.getPen()
        for operator, args in [
            ("moveTo", ((0, 0),)),
            ("lineTo", ((1, 1),)),
            ("endPath", ()),
        ]:
            getattr(pen1, operator)(*args)

        font2 = ufoLib2.Font()
        font2.info.unitsPerEm = 1000
        glyph2 = font2.newGlyph("a")
        pen2 = glyph2.getPen()
        for operator, args in [
            ("moveTo", ((0, 0),)),
            ("curveTo", ((1, 1), (2, 2), (3, 3))),
            ("endPath", ()),
        ]:
            getattr(pen2, operator)(*args)

        with pytest.raises(IncompatibleFontsError) as excinfo:
            fonts_to_quadratic([font1, font2])
        assert excinfo.match("fonts contains incompatible glyphs: 'a'")

        assert hasattr(excinfo.value, "glyph_errors")
        error = excinfo.value.glyph_errors["a"]
        assert isinstance(error, IncompatibleSegmentTypesError)
        assert error.segments == {1: ["line", "curve"]}

    def test_already_quadratic(self):
        glyph = ufoLib2.objects.Glyph()
        pen = glyph.getPen()
        pen.moveTo((0, 0))
        pen.qCurveTo((1, 1), (2, 2))
        pen.closePath()
        assert not glyph_to_quadratic(glyph)

    def test_open_paths(self):
        glyph = ufoLib2.objects.Glyph()
        pen = glyph.getPen()
        pen.moveTo((0, 0))
        pen.lineTo((1, 1))
        pen.curveTo((2, 2), (3, 3), (4, 4))
        pen.endPath()
        assert glyph_to_quadratic(glyph)
        # open contour is still open
        assert glyph[-1][0].segmentType == "move"

    def test_ignore_components(self):
        glyph = ufoLib2.objects.Glyph()
        pen = glyph.getPen()
        pen.addComponent("a", (1, 0, 0, 1, 0, 0))
        pen.moveTo((0, 0))
        pen.curveTo((1, 1), (2, 2), (3, 3))
        pen.closePath()
        assert glyph_to_quadratic(glyph)
        assert len(glyph.components) == 1

    def test_overlapping_start_end_points(self):
        # https://github.com/googlefonts/fontmake/issues/572
        glyph1 = ufoLib2.objects.Glyph()
        pen = glyph1.getPointPen()
        pen.beginPath()
        pen.addPoint((0, 651), segmentType="line")
        pen.addPoint((0, 101), segmentType="line")
        pen.addPoint((0, 101), segmentType="line")
        pen.addPoint((0, 651), segmentType="line")
        pen.endPath()

        glyph2 = ufoLib2.objects.Glyph()
        pen = glyph2.getPointPen()
        pen.beginPath()
        pen.addPoint((1, 651), segmentType="line")
        pen.addPoint((2, 101), segmentType="line")
        pen.addPoint((3, 101), segmentType="line")
        pen.addPoint((4, 651), segmentType="line")
        pen.endPath()

        glyphs = [glyph1, glyph2]

        assert glyphs_to_quadratic(glyphs, reverse_direction=True)

        assert [[(p.x, p.y) for p in glyph[0]] for glyph in glyphs] == [
            [
                (0, 651),
                (0, 651),
                (0, 101),
                (0, 101),
            ],
            [(1, 651), (4, 651), (3, 101), (2, 101)],
        ]



from fontTools.designspaceLib import DesignSpaceDocument
from fontTools.designspaceLib.statNames import StatNames, getStatNames

from .fixtures import datadir


def test_instance_getStatNames(datadir):
    doc = DesignSpaceDocument.fromfile(datadir / "test_v5_sourceserif.designspace")

    assert getStatNames(doc, doc.instances[0].getFullUserLocation(doc)) == StatNames(
        familyNames={"en": "Source Serif 4"},
        styleNames={"en": "Caption ExtraLight"},
        postScriptFontName="SourceSerif4-CaptionExtraLight",
        styleMapFamilyNames={"en": "Source Serif 4 Caption ExtraLight"},
        styleMapStyleName="regular",
    )


def test_not_all_ordering_specified_and_translations(datadir):
    doc = DesignSpaceDocument.fromfile(datadir / "test_v5.designspace")

    assert getStatNames(doc, {"Weight": 200, "Width": 125, "Italic": 1}) == StatNames(
        familyNames={
            "en": "MasterFamilyName",
            "fr": "Montserrat",
            "ja": "モンセラート",
        },
        styleNames={
            "fr": "Wide Extra léger Italic",
            "de": "Wide Extraleicht Italic",
            "en": "Wide Extra Light Italic",
        },
        postScriptFontName="MasterFamilyName-WideExtraLightItalic",
        styleMapFamilyNames={
            "en": "MasterFamilyName Wide Extra Light",
            "fr": "Montserrat Wide Extra léger",
            "de": "MasterFamilyName Wide Extraleicht",
            "ja": "モンセラート Wide Extra Light",
        },
        styleMapStyleName="italic",
    )


def test_detect_ribbi_aktiv(datadir):
    doc = DesignSpaceDocument.fromfile(datadir / "test_v5_aktiv.designspace")

    assert getStatNames(doc, {"Weight": 600, "Width": 125, "Italic": 1}) == StatNames(
        familyNames={"en": "Aktiv Grotesk"},
        styleNames={"en": "Ex SemiBold Italic"},
        postScriptFontName="AktivGrotesk-ExSemiBoldItalic",
        styleMapFamilyNames={"en": "Aktiv Grotesk Ex SemiBold"},
        styleMapStyleName="italic",
    )

    assert getStatNames(doc, {"Weight": 700, "Width": 75, "Italic": 1}) == StatNames(
        familyNames={"en": "Aktiv Grotesk"},
        styleNames={"en": "Cd Bold Italic"},
        postScriptFontName="AktivGrotesk-CdBoldItalic",
        styleMapFamilyNames={"en": "Aktiv Grotesk Cd"},
        styleMapStyleName="bold italic",
    )


def test_detect_ribbi_recursive(datadir):
    doc = DesignSpaceDocument.fromfile(datadir / "test_v5.designspace")

    assert getStatNames(doc, {"Weight": 700, "Width": 125, "Italic": 1}) == StatNames(
        familyNames={
            "en": "MasterFamilyName",
            "fr": "Montserrat",
            "ja": "モンセラート",
        },
        styleNames={
            "en": "Wide Bold Italic",
        },
        postScriptFontName="MasterFamilyName-WideBoldItalic",
        styleMapFamilyNames={
            "en": "MasterFamilyName Wide",
            "fr": "Montserrat Wide",
            "ja": "モンセラート Wide",
        },
        styleMapStyleName="bold italic",
    )


def test_getStatNames_on_ds4_doesnt_make_up_bad_names(datadir):
    """See this issue on GitHub: https://github.com/googlefonts/ufo2ft/issues/630

    When as in the example, there's no STAT data present, the getStatName
    shouldn't try making up a postscript name.
    """
    doc = DesignSpaceDocument.fromfile(datadir / "DS5BreakTest.designspace")

    assert getStatNames(doc, {"Weight": 600, "Width": 125, "Italic": 1}) == StatNames(
        familyNames={"en": "DS5BreakTest"},
        styleNames={},
        postScriptFontName=None,
        styleMapFamilyNames={},
        styleMapStyleName=None,
    )



from fontTools.misc.testTools import FakeFont, getXML, parseXML
from fontTools.misc.textTools import deHexStr, hexStr
from fontTools.ttLib import newTable
import unittest


# On macOS X 10.12.3, the font /Library/Fonts/AppleGothic.ttf has a ‘gcid’
# table with a similar structure as this test data, just more CIDs.
GCID_DATA = deHexStr(
    "0000 0000 "  #   0: Format=0, Flags=0
    "0000 0098 "  #   4: Size=152
    "0000 "  #   8: Registry=0
    "41 64 6F 62 65 "  #  10: RegistryName="Adobe"
    + ("00" * 59)
    + "0003 "  #  15: <padding>  #  74: Order=3
    "4B 6F 72 65 61 31 "  #  76: Order="Korea1"
    + ("00" * 58)
    + "0001 "  #  82: <padding>  # 140: SupplementVersion
    "0004 "  # 142: Count
    "1234 "  # 144: CIDs[0/.notdef]=4660
    "FFFF "  # 146: CIDs[1/A]=None
    "0007 "  # 148: CIDs[2/B]=7
    "DEF0 "  # 150: CIDs[3/C]=57072
)  # 152: <end>
assert len(GCID_DATA) == 152, len(GCID_DATA)


GCID_XML = [
    '<GlyphCIDMapping Format="0">',
    '  <DataFormat value="0"/>',
    "  <!-- StructLength=152 -->",
    '  <Registry value="0"/>',
    '  <RegistryName value="Adobe"/>',
    '  <Order value="3"/>',
    '  <OrderName value="Korea1"/>',
    '  <SupplementVersion value="1"/>',
    "  <Mapping>",
    '    <CID glyph=".notdef" value="4660"/>',
    '    <CID glyph="B" value="7"/>',
    '    <CID glyph="C" value="57072"/>',
    "  </Mapping>",
    "</GlyphCIDMapping>",
]


class GCIDTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.maxDiff = None
        cls.font = FakeFont([".notdef", "A", "B", "C", "D"])

    def testDecompileToXML(self):
        table = newTable("gcid")
        table.decompile(GCID_DATA, self.font)
        self.assertEqual(getXML(table.toXML, self.font), GCID_XML)

    def testCompileFromXML(self):
        table = newTable("gcid")
        for name, attrs, content in parseXML(GCID_XML):
            table.fromXML(name, attrs, content, font=self.font)
        self.assertEqual(hexStr(table.compile(self.font)), hexStr(GCID_DATA))


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



import logging
import pytest
from pathlib import Path

pathops = pytest.importorskip("pathops")

from fontTools.ttLib import TTFont
from fontTools.ttLib.removeOverlaps import removeOverlaps, _simplify, _round_path

DATA_DIR = Path(__file__).parent / "data"


def test_pathops_simplify_bug_workaround(caplog):
    # Paths extracted from Noto Sans Ethiopic instance that fails skia-pathops
    # https://github.com/google/fonts/issues/3365
    # https://bugs.chromium.org/p/skia/issues/detail?id=11958
    path = pathops.Path()
    path.moveTo(550.461, 0)
    path.lineTo(550.461, 366.308)
    path.lineTo(713.229, 366.308)
    path.lineTo(713.229, 0)
    path.close()
    path.moveTo(574.46, 0)
    path.lineTo(574.46, 276.231)
    path.lineTo(737.768, 276.231)
    path.quadTo(820.075, 276.231, 859.806, 242.654)
    path.quadTo(899.537, 209.077, 899.537, 144.154)
    path.quadTo(899.537, 79, 853.46, 39.5)
    path.quadTo(807.383, 0, 712.383, 0)
    path.close()

    # check that it fails without workaround
    with pytest.raises(pathops.PathOpsError):
        pathops.simplify(path)

    # check our workaround works (but with a warning)
    with caplog.at_level(logging.DEBUG, logger="fontTools.ttLib.removeOverlaps"):
        result = _simplify(path, debugGlyphName="a")

    assert "skia-pathops failed to simplify 'a' with float coordinates" in caplog.text

    expected = pathops.Path()
    expected.moveTo(550, 0)
    expected.lineTo(550, 366)
    expected.lineTo(713, 366)
    expected.lineTo(713, 276)
    expected.lineTo(738, 276)
    expected.quadTo(820, 276, 860, 243)
    expected.quadTo(900, 209, 900, 144)
    expected.quadTo(900, 79, 853, 40)
    expected.quadTo(807.242, 0.211, 713, 0.001)
    expected.lineTo(713, 0)
    expected.close()

    assert expected == _round_path(result, round=lambda v: round(v, 3))


def test_CFF_CharString_width_nominalWidthX():
    font_path = DATA_DIR / "IBMPlexSans-Bold.subset.otf"
    font = TTFont(str(font_path))

    assert font["hmtx"]["OE"][0] == 998

    # calcBounds() has the side effect of setting the width attribute
    font["CFF "].cff[0].CharStrings["OE"].calcBounds({})
    assert font["CFF "].cff[0].CharStrings["OE"].width == font["hmtx"]["OE"][0]

    removeOverlaps(font)

    assert font["hmtx"]["OE"][0] == 998

    font["CFF "].cff[0].CharStrings["OE"].calcBounds({})
    assert font["CFF "].cff[0].CharStrings["OE"].width == font["hmtx"]["OE"][0]



from fontTools.cffLib import maxStackLimit as maxStack
from fontTools.cffLib.specializer import (
    programToString,
    stringToProgram,
    generalizeProgram,
    specializeProgram,
    programToCommands,
    commandsToProgram,
    generalizeCommands,
    specializeCommands,
)
from fontTools.ttLib import TTFont
import os
import pytest
from fontTools.misc.testTools import parseXML, DataFilesHandler


def charstr_generalize(charstr, **kwargs):
    return programToString(generalizeProgram(stringToProgram(charstr), **kwargs))


def charstr_specialize(charstr, **kwargs):
    return programToString(specializeProgram(stringToProgram(charstr), **kwargs))


def program_stack_use(program, getNumRegions=None):
    vsindex = None
    maxStack = 0
    stack = []
    for token in program:
        if token == vsindex:
            vsindex = stack[-1]
            stack = []
        elif token == "blend":
            numBlends = stack[-1]
            numRegions = getNumRegions(vsindex)
            numOperandsToPop = 1 + numBlends * numRegions
            stack[-numOperandsToPop:] = []
        elif type(token) is str:
            stack = []
        else:
            stack.append(token)
            maxStack = max(maxStack, len(stack))

    return maxStack


class CFFGeneralizeProgramTest:
    @pytest.mark.parametrize(
        "charstr",
        [
            # no arguments/operands
            ("rmoveto"),  # test_rmoveto_none
            ("hmoveto"),  # test_hmoveto_none
            ("vmoveto"),  # test_vmoveto_none
            ("rlineto"),  # test_rlineto_none
            ("hlineto"),  # test_hlineto_none
            ("vlineto"),  # test_vlineto_none
            ("rrcurveto"),  # test_rrcurveto_none
            ("hhcurveto"),  # test_hhcurveto_none
            ("vvcurveto"),  # test_vvcurveto_none
            ("hvcurveto"),  # test_hvcurveto_none
            ("vhcurveto"),  # test_vhcurveto_none
            ("rcurveline"),  # test_rcurveline_none
            ("rlinecurve"),  # test_rlinecurve_none
        ],
    )
    def test_raises(self, charstr):
        try:
            charstr_generalize(charstr)
        except ValueError:
            return
        raise AssertionError("Expected to raise ValueError, but didn't.", charstr)

    @pytest.mark.parametrize(
        "charstr, expected",
        [
            # rmoveto
            ("0 0 rmoveto", None),  # test_rmoveto_zero
            ("100 0 0 rmoveto", None),  # test_rmoveto_zero_width
            (".55 -.8 rmoveto", "0.55 -0.8 rmoveto"),  # test_rmoveto
            ("100.5 50 -5.8 rmoveto", None),  # test_rmoveto_width
            # hmoveto
            ("0 hmoveto", "0 0 rmoveto"),  # test_hmoveto_zero
            ("100 0 hmoveto", "100 0 0 rmoveto"),  # test_hmoveto_zero_width
            (".67 hmoveto", "0.67 0 rmoveto"),  # test_hmoveto
            ("100 -70 hmoveto", "100 -70 0 rmoveto"),  # test_hmoveto_width
            # vmoveto
            ("0 vmoveto", "0 0 rmoveto"),  # test_vmoveto_zero
            ("100 0 vmoveto", "100 0 0 rmoveto"),  # test_vmoveto_zero_width
            ("-.24 vmoveto", "0 -0.24 rmoveto"),  # test_vmoveto
            ("100 44 vmoveto", "100 0 44 rmoveto"),  # test_vmoveto_width
            # rlineto
            ("0 0 rlineto", None),  # test_rlineto_zero
            (  # test_rlineto_zero_mult
                "0 0 0 0 0 0 rlineto",
                "0 0 rlineto " * 3,
            ),
            (".55 -.8 rlineto", "0.55 -0.8 rlineto"),  # test_rlineto
            (  # test_rlineto_mult
                ".55 -.8 .55 -.8 .55 -.8 rlineto",
                "0.55 -0.8 rlineto " * 3,
            ),
            # hlineto
            ("0 hlineto", "0 0 rlineto"),  # test_hlineto_zero
            (  # test_hlineto_zero_mult
                "0 0 0 0 hlineto",
                "0 0 rlineto " * 4,
            ),
            (".67 hlineto", "0.67 0 rlineto"),  # test_hlineto
            (  # test_hlineto_mult
                ".67 -6.0 .67 hlineto",
                "0.67 0 rlineto 0 -6.0 rlineto 0.67 0 rlineto",
            ),
            # vlineto
            ("0 vlineto", "0 0 rlineto"),  # test_vlineto_zero
            ("0 0 0 vlineto", "0 0 rlineto " * 3),  # test_vlineto_zero_mult
            ("-.24 vlineto", "0 -0.24 rlineto"),  # test_vlineto
            (  # test_vlineto_mult
                "-.24 +50 30 -4 vlineto",
                "0 -0.24 rlineto 50 0 rlineto 0 30 rlineto -4 0 rlineto",
            ),
            # rrcurveto
            ("-1 56 -2 57 -1 57 rrcurveto", None),  # test_rrcurveto
            (  # test_rrcurveto_mult
                "-30 8 -36 15 -37 22 44 54 31 61 22 68 rrcurveto",
                "-30 8 -36 15 -37 22 rrcurveto 44 54 31 61 22 68 rrcurveto",
            ),
            ("1 2 3 4 5 0 rrcurveto", None),  # test_rrcurveto_d3947b8
            (  # test_rrcurveto_v0_0h_h0
                "0 10 1 2 0 0 0 0 1 2 0 1 0 1 3 4 0 0 rrcurveto",
                "0 10 1 2 0 0 rrcurveto 0 0 1 2 0 1 rrcurveto 0 1 3 4 0 0 rrcurveto",
            ),
            (  # test_rrcurveto_h0_0h_h0
                "10 0 1 2 0 0 0 0 1 2 0 1 0 1 3 4 0 0 rrcurveto",
                "10 0 1 2 0 0 rrcurveto 0 0 1 2 0 1 rrcurveto 0 1 3 4 0 0 rrcurveto",
            ),
            (  # test_rrcurveto_00_0h_h0
                "0 0 1 2 0 0 0 0 1 2 0 1 0 1 3 4 0 0 rrcurveto",
                "0 0 1 2 0 0 rrcurveto 0 0 1 2 0 1 rrcurveto 0 1 3 4 0 0 rrcurveto",
            ),
            (  # test_rrcurveto_r0_0h_h0
                "10 10 1 2 0 0 0 0 1 2 0 1 0 1 3 4 0 0 rrcurveto",
                "10 10 1 2 0 0 rrcurveto 0 0 1 2 0 1 rrcurveto 0 1 3 4 0 0 rrcurveto",
            ),
            (  # test_rrcurveto_v0_0v_v0
                "0 10 1 2 0 0 0 0 1 2 1 0 1 0 3 4 0 0 rrcurveto",
                "0 10 1 2 0 0 rrcurveto 0 0 1 2 1 0 rrcurveto 1 0 3 4 0 0 rrcurveto",
            ),
            (  # test_rrcurveto_h0_0v_v0
                "10 0 1 2 0 0 0 0 1 2 1 0 1 0 3 4 0 0 rrcurveto",
                "10 0 1 2 0 0 rrcurveto 0 0 1 2 1 0 rrcurveto 1 0 3 4 0 0 rrcurveto",
            ),
            (  # test_rrcurveto_00_0v_v0
                "0 0 1 2 0 0 0 0 1 2 1 0 1 0 3 4 0 0 rrcurveto",
                "0 0 1 2 0 0 rrcurveto 0 0 1 2 1 0 rrcurveto 1 0 3 4 0 0 rrcurveto",
            ),
            (  # test_rrcurveto_r0_0v_v0
                "10 10 1 2 0 0 0 0 1 2 1 0 1 0 3 4 0 0 rrcurveto",
                "10 10 1 2 0 0 rrcurveto 0 0 1 2 1 0 rrcurveto 1 0 3 4 0 0 rrcurveto",
            ),
            # hhcurveto
            ("10 30 0 10 hhcurveto", "10 0 30 0 10 0 rrcurveto"),  # test_hhcurveto_4
            (  # test_hhcurveto_5
                "40 -38 -60 41 -91 hhcurveto",
                "-38 40 -60 41 -91 0 rrcurveto",
            ),
            (  # test_hhcurveto_mult_4_4
                "43 23 25 18 29 56 42 -84 hhcurveto",
                "43 0 23 25 18 0 rrcurveto 29 0 56 42 -84 0 rrcurveto",
            ),
            (  # test_hhcurveto_mult_5_4
                "43 23 25 18 29 56 42 -84 79 hhcurveto",
                "23 43 25 18 29 0 rrcurveto 56 0 42 -84 79 0 rrcurveto",
            ),
            (  # test_hhcurveto_mult_4_4_4
                "1 2 3 4 5 6 7 8 9 10 11 12 hhcurveto",
                "1 0 2 3 4 0 rrcurveto 5 0 6 7 8 0 rrcurveto 9 0 10 11 12 0 rrcurveto",
            ),
            (  # test_hhcurveto_mult_5_4_4
                "1 2 3 4 5 6 7 8 9 10 11 12 13 hhcurveto",
                "2 1 3 4 5 0 rrcurveto 6 0 7 8 9 0 rrcurveto 10 0 11 12 13 0 rrcurveto",
            ),
            # vvcurveto
            ("61 6 52 68 vvcurveto", "0 61 6 52 0 68 rrcurveto"),  # test_vvcurveto_4
            (  # test_vvcurveto_5
                "61 38 35 56 72 vvcurveto",
                "61 38 35 56 0 72 rrcurveto",
            ),
            (  # test_vvcurveto_mult_4_4
                "-84 -88 -30 -90 -13 19 23 -11 vvcurveto",
                "0 -84 -88 -30 0 -90 rrcurveto 0 -13 19 23 0 -11 rrcurveto",
            ),
            (  # test_vvcurveto_mult_5_4
                "43 12 17 32 65 68 -6 52 61 vvcurveto",
                "43 12 17 32 0 65 rrcurveto 0 68 -6 52 0 61 rrcurveto",
            ),
            (  # test_vvcurveto_mult_4_4_4
                "1 2 3 4 5 6 7 8 9 10 11 12 vvcurveto",
                "0 1 2 3 0 4 rrcurveto 0 5 6 7 0 8 rrcurveto 0 9 10 11 0 12 rrcurveto",
            ),
            (  # test_vvcurveto_mult_5_4_4
                "1 2 3 4 5 6 7 8 9 10 11 12 13 vvcurveto",
                "1 2 3 4 0 5 rrcurveto 0 6 7 8 0 9 rrcurveto 0 10 11 12 0 13 rrcurveto",
            ),
            # hvcurveto
            ("1 2 3 4 hvcurveto", "1 0 2 3 0 4 rrcurveto"),  # test_hvcurveto_4
            (  # test_hvcurveto_5
                "57 44 22 40 34 hvcurveto",
                "57 0 44 22 34 40 rrcurveto",
            ),
            (  # test_hvcurveto_4_4
                "65 33 -19 -45 -45 -29 -25 -71 hvcurveto",
                "65 0 33 -19 0 -45 rrcurveto 0 -45 -29 -25 -71 0 rrcurveto",
            ),
            (  # test_hvcurveto_4_5
                "97 69 41 86 58 -36 34 -64 11 hvcurveto",
                "97 0 69 41 0 86 rrcurveto 0 58 -36 34 -64 11 rrcurveto",
            ),
            (  # test_hvcurveto_4_4_4
                "1 2 3 4 5 6 7 8 9 10 11 12 hvcurveto",
                "1 0 2 3 0 4 rrcurveto 0 5 6 7 8 0 rrcurveto 9 0 10 11 0 12 rrcurveto",
            ),
            (  # test_hvcurveto_4_4_5
                "-124 -79 104 165 163 82 102 124 56 43 -25 -37 35 hvcurveto",
                "-124 0 -79 104 0 165 rrcurveto 0 163 82 102 124 0 rrcurveto 56 0 43 -25 35 -37 rrcurveto",
            ),
            (  # test_hvcurveto_4_4_4_4
                "32 25 22 32 31 -25 22 -32 -32 -25 -22 -31 -32 25 -22 32 hvcurveto",
                "32 0 25 22 0 32 rrcurveto 0 31 -25 22 -32 0 rrcurveto -32 0 -25 -22 0 -31 rrcurveto 0 -32 25 -22 32 0 rrcurveto",
            ),
            (  # test_hvcurveto_4_4_4_4_5
                "-170 -128 111 195 234 172 151 178 182 95 -118 -161 -130 -71 -77 -63 -55 -19 38 79 20 hvcurveto",
                "-170 0 -128 111 0 195 rrcurveto 0 234 172 151 178 0 rrcurveto 182 0 95 -118 0 -161 rrcurveto 0 -130 -71 -77 -63 0 rrcurveto -55 0 -19 38 20 79 rrcurveto",
            ),
            # vhcurveto
            (  # test_vhcurveto_4
                "-57 43 -30 53 vhcurveto",
                "0 -57 43 -30 53 0 rrcurveto",
            ),
            (  # test_vhcurveto_5
                "41 -27 19 -46 11 vhcurveto",
                "0 41 -27 19 -46 11 rrcurveto",
            ),
            (  # test_vhcurveto_4_4
                "1 2 3 4 5 6 7 8 vhcurveto",
                "0 1 2 3 4 0 rrcurveto 5 0 6 7 0 8 rrcurveto",
            ),
            (  # test_vhcurveto_4_5
                "-64 -23 -25 -45 -30 -24 14 33 -19 vhcurveto",
                "0 -64 -23 -25 -45 0 rrcurveto -30 0 -24 14 -19 33 rrcurveto",
            ),
            (  # test_vhcurveto_4_4_4
                "1 2 3 4 5 6 7 8 9 10 11 12 vhcurveto",
                "0 1 2 3 4 0 rrcurveto 5 0 6 7 0 8 rrcurveto 0 9 10 11 12 0 rrcurveto",
            ),
            (  # test_vhcurveto_4_4_5
                "108 59 81 98 99 59 -81 -108 -100 -46 -66 -63 -47 vhcurveto",
                "0 108 59 81 98 0 rrcurveto 99 0 59 -81 0 -108 rrcurveto 0 -100 -46 -66 -63 -47 rrcurveto",
            ),
            (  # test_vhcurveto_4_4_4_5
                "60 -26 37 -43 -33 -28 -22 -36 -37 27 -20 32 3 4 0 1 3 vhcurveto",
                "0 60 -26 37 -43 0 rrcurveto -33 0 -28 -22 0 -36 rrcurveto 0 -37 27 -20 32 0 rrcurveto 3 0 4 0 3 1 rrcurveto",
            ),
            # rcurveline
            (  # test_rcurveline_6_2
                "21 -76 21 -72 24 -73 31 -100 rcurveline",
                "21 -76 21 -72 24 -73 rrcurveto 31 -100 rlineto",
            ),
            (  # test_rcurveline_6_6_2
                "-73 80 -80 121 -49 96 60 65 55 41 54 17 -8 78 rcurveline",
                "-73 80 -80 121 -49 96 rrcurveto 60 65 55 41 54 17 rrcurveto -8 78 rlineto",
            ),
            (  # test_rcurveline_6_6_6_2
                "1 64 10 51 29 39 15 21 15 20 15 18 47 -89 63 -98 52 -59 91 8 rcurveline",
                "1 64 10 51 29 39 rrcurveto 15 21 15 20 15 18 rrcurveto 47 -89 63 -98 52 -59 rrcurveto 91 8 rlineto",
            ),
            (  # test_rcurveline_6_6_6_6_2
                "1 64 10 51 29 39 15 21 15 20 15 18 46 -88 63 -97 52 -59 -38 -57 -49 -62 -52 -54 96 -8 rcurveline",
                "1 64 10 51 29 39 rrcurveto 15 21 15 20 15 18 rrcurveto 46 -88 63 -97 52 -59 rrcurveto -38 -57 -49 -62 -52 -54 rrcurveto 96 -8 rlineto",
            ),
            # rlinecurve
            (  # test_rlinecurve_2_6
                "21 -76 21 -72 24 -73 31 -100 rlinecurve",
                "21 -76 rlineto 21 -72 24 -73 31 -100 rrcurveto",
            ),
            (  # test_rlinecurve_2_2_6
                "-73 80 -80 121 -49 96 60 65 55 41 rlinecurve",
                "-73 80 rlineto -80 121 rlineto -49 96 60 65 55 41 rrcurveto",
            ),
            (  # test_rlinecurve_2_2_2_6
                "1 64 10 51 29 39 15 21 15 20 15 18 rlinecurve",
                "1 64 rlineto 10 51 rlineto 29 39 rlineto 15 21 15 20 15 18 rrcurveto",
            ),
            (  # test_rlinecurve_2_2_2_2_6
                "1 64 10 51 29 39 15 21 15 20 15 18 46 -88 rlinecurve",
                "1 64 rlineto 10 51 rlineto 29 39 rlineto 15 21 rlineto 15 20 15 18 46 -88 rrcurveto",
            ),
            # hstem/vstem
            (  # test_hstem_vstem
                "95 0 58 542 60 hstem 89 65 344 67 vstem 89 45 rmoveto",
                None,
            ),
            # hstemhm/vstemhm
            (  # test_hstemhm_vstemhm
                "-16 577 60 24 60 hstemhm 98 55 236 55 vstemhm 343 577 rmoveto",
                None,
            ),
            # hintmask/cntrmask
            (  # test_hintmask_cntrmask
                "52 80 153 61 4 83 -71.5 71.5 hintmask 11011100 94 119 216 119 216 119 cntrmask 1110000 154 -12 rmoveto",
                None,
            ),
            # endchar
            ("-255 319 rmoveto 266 57 rlineto endchar", None),  # test_endchar
            # xtra
            ("-255 319 rmoveto 266 57 rlineto xtra 90 34", None),  # test_xtra
        ],
    )
    def test_generalize(self, charstr, expected):
        if expected is None:
            expected = charstr
        expected = expected.strip()
        generalized = charstr_generalize(charstr)
        assert generalized == expected, (generalized, expected)


class CFFSpecializeProgramTest:
    @pytest.mark.parametrize(
        "charstr",
        [
            # no arguments/operands
            ("rmoveto"),  # test_rmoveto_none
            ("hmoveto"),  # test_hmoveto_none
            ("vmoveto"),  # test_vmoveto_none
            ("rlineto"),  # test_rlineto_none
            ("hlineto"),  # test_hlineto_none
            ("vlineto"),  # test_vlineto_none
            ("rrcurveto"),  # test_rrcurveto_none
            ("hhcurveto"),  # test_hhcurveto_none
            ("vvcurveto"),  # test_vvcurveto_none
            ("hvcurveto"),  # test_hvcurveto_none
            ("vhcurveto"),  # test_vhcurveto_none
            ("rcurveline"),  # test_rcurveline_none
            ("rlinecurve"),  # test_rlinecurve_none
        ],
    )
    def test_raises(self, charstr):
        try:
            charstr_specialize(charstr)
        except ValueError:
            return
        raise AssertionError("Expected to raise ValueError, but didn't.", charstr)

    @pytest.mark.parametrize(
        "charstr, expected",
        [
            # rmoveto
            ("0 0 rmoveto", "0 hmoveto"),  # test_rmoveto_zero
            ("0 0 rmoveto " * 3, "0 hmoveto"),  # test_rmoveto_zero_mult
            ("100 0 0 rmoveto", "100 0 hmoveto"),  # test_rmoveto_zero_width
            (".55 -.8 rmoveto", "0.55 -0.8 rmoveto"),  # test_rmoveto
            ("55 -8 rmoveto " * 3, "165 -24 rmoveto"),  # test_rmoveto_mult
            ("100.5 50 -5.8 rmoveto", None),  # test_rmoveto_width
            # rlineto
            ("0 0 rlineto", ""),  # test_rlineto_zero
            ("0 0 rlineto " * 3, ""),  # test_rlineto_zero_mult
            (".55 -.8 rlineto", "0.55 -0.8 rlineto"),  # test_rlineto
            (  # test_rlineto_mult
                ".55 -.8 rlineto " * 3,
                "0.55 -0.8 0.55 -0.8 0.55 -0.8 rlineto",
            ),
            (".67 0 rlineto", "0.67 hlineto"),  # test_hlineto
            ("62 0 rlineto " * 3, "186 hlineto"),  # test_hlineto_zero_mult
            (  # test_hlineto_mult
                ".67 0 rlineto 0 -6.0 rlineto .67 0 rlineto",
                "0.67 -6.0 0.67 hlineto",
            ),
            ("0 -.24 rlineto", "-0.24 vlineto"),  # test_vlineto
            ("0 -24 rlineto " * 3, "-72 vlineto"),  # test_vlineto_zero_mult
            (  # test_vlineto_mult
                "0 -.24 rlineto +50 0 rlineto 0 30 rlineto -4 0 rlineto",
                "-0.24 50 30 -4 vlineto",
            ),
            ("1 2 0 0 3 4 rlineto", "1 2 3 4 rlineto"),  # test_0lineto_peephole
            ("1 2 5 0 3 4 rlineto", None),  # test_hlineto_peephole
            ("1 2 0 5 3 4 rlineto", None),  # test_vlineto_peephole
            # rrcurveto
            ("-1 56 -2 57 -1 57 rrcurveto", None),  # test_rrcurveto
            (  # test_rrcurveto_mult
                "-30 8 -36 15 -37 22 rrcurveto 44 54 31 61 22 68 rrcurveto",
                "-30 8 -36 15 -37 22 44 54 31 61 22 68 rrcurveto",
            ),
            ("1 2 3 4 5 0 rrcurveto", "2 1 3 4 5 hhcurveto"),  # test_rrcurveto_d3947b8
            ("10 0 30 0 10 0 rrcurveto", "10 30 0 10 hhcurveto"),  # test_hhcurveto_4
            (  # test_hhcurveto_5
                "-38 40 -60 41 -91 0 rrcurveto",
                "40 -38 -60 41 -91 hhcurveto",
            ),
            (  # test_hhcurveto_mult_4_4
                "43 0 23 25 18 0 rrcurveto 29 0 56 42 -84 0 rrcurveto",
                "43 23 25 18 29 56 42 -84 hhcurveto",
            ),
            (  # test_hhcurveto_mult_5_4
                "23 43 25 18 29 0 rrcurveto 56 0 42 -84 79 0 rrcurveto",
                "43 23 25 18 29 56 42 -84 79 hhcurveto",
            ),
            (  # test_hhcurveto_mult_4_4_4
                "1 0 2 3 4 0 rrcurveto 5 0 6 7 8 0 rrcurveto 9 0 10 11 12 0 rrcurveto",
                "1 2 3 4 5 6 7 8 9 10 11 12 hhcurveto",
            ),
            (  # test_hhcurveto_mult_5_4_4
                "2 1 3 4 5 0 rrcurveto 6 0 7 8 9 0 rrcurveto 10 0 11 12 13 0 rrcurveto",
                "1 2 3 4 5 6 7 8 9 10 11 12 13 hhcurveto",
            ),
            ("0 61 6 52 0 68 rrcurveto", "61 6 52 68 vvcurveto"),  # test_vvcurveto_4
            (  # test_vvcurveto_5
                "61 38 35 56 0 72 rrcurveto",
                "61 38 35 56 72 vvcurveto",
            ),
            (  # test_vvcurveto_mult_4_4
                "0 -84 -88 -30 0 -90 rrcurveto 0 -13 19 23 0 -11 rrcurveto",
                "-84 -88 -30 -90 -13 19 23 -11 vvcurveto",
            ),
            (  # test_vvcurveto_mult_5_4
                "43 12 17 32 0 65 rrcurveto 0 68 -6 52 0 61 rrcurveto",
                "43 12 17 32 65 68 -6 52 61 vvcurveto",
            ),
            (  # test_vvcurveto_mult_4_4_4
                "0 1 2 3 0 4 rrcurveto 0 5 6 7 0 8 rrcurveto 0 9 10 11 0 12 rrcurveto",
                "1 2 3 4 5 6 7 8 9 10 11 12 vvcurveto",
            ),
            (  # test_vvcurveto_mult_5_4_4
                "1 2 3 4 0 5 rrcurveto 0 6 7 8 0 9 rrcurveto 0 10 11 12 0 13 rrcurveto",
                "1 2 3 4 5 6 7 8 9 10 11 12 13 vvcurveto",
            ),
            ("1 0 2 3 0 4 rrcurveto", "1 2 3 4 hvcurveto"),  # test_hvcurveto_4
            (  # test_hvcurveto_5
                "57 0 44 22 34 40 rrcurveto",
                "57 44 22 40 34 hvcurveto",
            ),
            (  # test_hvcurveto_4_4
                "65 0 33 -19 0 -45 rrcurveto 0 -45 -29 -25 -71 0 rrcurveto",
                "65 33 -19 -45 -45 -29 -25 -71 hvcurveto",
            ),
            (  # test_hvcurveto_4_5
                "97 0 69 41 0 86 rrcurveto 0 58 -36 34 -64 11 rrcurveto",
                "97 69 41 86 58 -36 34 -64 11 hvcurveto",
            ),
            (  # test_hvcurveto_4_4_4
                "1 0 2 3 0 4 rrcurveto 0 5 6 7 8 0 rrcurveto 9 0 10 11 0 12 rrcurveto",
                "1 2 3 4 5 6 7 8 9 10 11 12 hvcurveto",
            ),
            (  # test_hvcurveto_4_4_5
                "-124 0 -79 104 0 165 rrcurveto 0 163 82 102 124 0 rrcurveto 56 0 43 -25 35 -37 rrcurveto",
                "-124 -79 104 165 163 82 102 124 56 43 -25 -37 35 hvcurveto",
            ),
            (  # test_hvcurveto_4_4_4_4
                "32 0 25 22 0 32 rrcurveto 0 31 -25 22 -32 0 rrcurveto -32 0 -25 -22 0 -31 rrcurveto 0 -32 25 -22 32 0 rrcurveto",
                "32 25 22 32 31 -25 22 -32 -32 -25 -22 -31 -32 25 -22 32 hvcurveto",
            ),
            (  # test_hvcurveto_4_4_4_4_5
                "-170 0 -128 111 0 195 rrcurveto 0 234 172 151 178 0 rrcurveto 182 0 95 -118 0 -161 rrcurveto 0 -130 -71 -77 -63 0 rrcurveto -55 0 -19 38 20 79 rrcurveto",
                "-170 -128 111 195 234 172 151 178 182 95 -118 -161 -130 -71 -77 -63 -55 -19 38 79 20 hvcurveto",
            ),
            (  # test_vhcurveto_4
                "0 -57 43 -30 53 0 rrcurveto",
                "-57 43 -30 53 vhcurveto",
            ),
            (  # test_vhcurveto_5
                "0 41 -27 19 -46 11 rrcurveto",
                "41 -27 19 -46 11 vhcurveto",
            ),
            (  # test_vhcurveto_4_4
                "0 1 2 3 4 0 rrcurveto 5 0 6 7 0 8 rrcurveto",
                "1 2 3 4 5 6 7 8 vhcurveto",
            ),
            (  # test_vhcurveto_4_5
                "0 -64 -23 -25 -45 0 rrcurveto -30 0 -24 14 -19 33 rrcurveto",
                "-64 -23 -25 -45 -30 -24 14 33 -19 vhcurveto",
            ),
            (  # test_vhcurveto_4_4_4
                "0 1 2 3 4 0 rrcurveto 5 0 6 7 0 8 rrcurveto 0 9 10 11 12 0 rrcurveto",
                "1 2 3 4 5 6 7 8 9 10 11 12 vhcurveto",
            ),
            (  # test_vhcurveto_4_4_5
                "0 108 59 81 98 0 rrcurveto 99 0 59 -81 0 -108 rrcurveto 0 -100 -46 -66 -63 -47 rrcurveto",
                "108 59 81 98 99 59 -81 -108 -100 -46 -66 -63 -47 vhcurveto",
            ),
            (  # test_vhcurveto_4_4_4_5
                "0 60 -26 37 -43 0 rrcurveto -33 0 -28 -22 0 -36 rrcurveto 0 -37 27 -20 32 0 rrcurveto 3 0 4 0 3 1 rrcurveto",
                "60 -26 37 -43 -33 -28 -22 -36 -37 27 -20 32 3 4 0 1 3 vhcurveto",
            ),
            (  # test_rrcurveto_v0_0h_h0
                "0 10 1 2 0 0 0 0 1 2 0 1 0 1 3 4 0 0 rrcurveto",
                "10 1 2 0 0 1 2 1 1 3 4 0 vhcurveto",
            ),
            (  # test_rrcurveto_h0_0h_h0
                "10 0 1 2 0 0 0 0 1 2 0 1 0 1 3 4 0 0 rrcurveto",
                "10 1 2 0 hhcurveto 0 1 2 1 1 3 4 0 hvcurveto",
            ),
            (  # test_rrcurveto_00_0h_h0
                "0 0 1 2 0 0 0 0 1 2 0 1 0 1 3 4 0 0 rrcurveto",
                "1 2 rlineto 0 1 2 1 1 3 4 0 hvcurveto",
            ),
            (  # test_rrcurveto_r0_0h_h0
                "10 10 1 2 0 0 0 0 1 2 0 1 0 1 3 4 0 0 rrcurveto",
                "10 10 1 2 0 0 1 2 1 1 3 4 0 vvcurveto",
            ),
            (  # test_rrcurveto_v0_0v_v0
                "0 10 1 2 0 0 0 0 1 2 1 0 1 0 3 4 0 0 rrcurveto",
                "10 1 2 0 vhcurveto 0 1 2 1 1 3 4 0 hhcurveto",
            ),
            (  # test_rrcurveto_h0_0v_v0
                "10 0 1 2 0 0 0 0 1 2 1 0 1 0 3 4 0 0 rrcurveto",
                "10 1 2 0 0 1 2 1 1 3 4 0 hhcurveto",
            ),
            (  # test_rrcurveto_00_0v_v0
                "0 0 1 2 0 0 0 0 1 2 1 0 1 0 3 4 0 0 rrcurveto",
                "1 2 rlineto 0 1 2 1 1 3 4 0 hhcurveto",
            ),
            (  # test_rrcurveto_r0_0v_v0
                "10 10 1 2 0 0 0 0 1 2 1 0 1 0 3 4 0 0 rrcurveto",
                "10 10 1 2 0 0 1 2 1 1 3 4 0 hhcurveto",
            ),
            (  # test_hhcurveto_peephole
                "1 2 3 4 5 6 1 2 3 4 5 0 1 2 3 4 5 6 rrcurveto",
                None,
            ),
            (  # test_vvcurveto_peephole
                "1 2 3 4 5 6 1 2 3 4 0 6 1 2 3 4 5 6 rrcurveto",
                None,
            ),
            (  # test_hvcurveto_peephole
                "1 2 3 4 5 6 1 0 3 4 5 6 1 2 3 4 5 6 rrcurveto",
                None,
            ),
            (  # test_vhcurveto_peephole
                "1 2 3 4 5 6 0 2 3 4 5 6 1 2 3 4 5 6 rrcurveto",
                None,
            ),
            (  # test_rcurveline_6_2
                "21 -76 21 -72 24 -73 rrcurveto 31 -100 rlineto",
                "21 -76 21 -72 24 -73 31 -100 rcurveline",
            ),
            (  # test_rcurveline_6_6_2
                "-73 80 -80 121 -49 96 rrcurveto 60 65 55 41 54 17 rrcurveto -8 78 rlineto",
                "-73 80 -80 121 -49 96 60 65 55 41 54 17 -8 78 rcurveline",
            ),
            (  # test_rcurveline_6_6_6_2
                "1 64 10 51 29 39 rrcurveto 15 21 15 20 15 18 rrcurveto 47 -89 63 -98 52 -59 rrcurveto 91 8 rlineto",
                "1 64 10 51 29 39 15 21 15 20 15 18 47 -89 63 -98 52 -59 91 8 rcurveline",
            ),
            (  # test_rlinecurve_2_6
                "21 -76 rlineto 21 -72 24 -73 31 -100 rrcurveto",
                "21 -76 21 -72 24 -73 31 -100 rlinecurve",
            ),
            (  # test_rlinecurve_2_2_6
                "-73 80 rlineto -80 121 rlineto -49 96 60 65 55 41 rrcurveto",
                "-73 80 -80 121 -49 96 60 65 55 41 rlinecurve",
            ),
            (  # test_rlinecurve_2_2_2_6
                "1 64 rlineto 10 51 rlineto 29 39 rlineto 15 21 15 20 15 18 rrcurveto",
                "1 64 10 51 29 39 15 21 15 20 15 18 rlinecurve",
            ),
        ],
    )
    def test_specialize(self, charstr, expected):
        if expected is None:
            expected = charstr
        expected = expected.strip()
        specialized = charstr_specialize(charstr)
        assert specialized == expected, (specialized, expected)

    # maxstack CFF=48, specializer uses up to 47
    def test_maxstack(self):
        operands = "1 2 3 4 5 6 "
        operator = "rrcurveto "
        charstr = (operands + operator) * 9
        expected = (operands * 2 + operator + operands * 7 + operator).rstrip()
        specialized = charstr_specialize(charstr)
        assert specialized == expected, (specialized, expected)

    # maxstack CFF2=513, specializer uses up to 512
    def test_maxstack_blends(self):
        numRegions = 15
        numOps = 600
        getNumRegions = lambda iv: numRegions
        blend_one = [i for i in range(1 + numRegions)] + [1, "blend"]
        operands = blend_one * 6
        operator = "rrcurveto"
        program = (operands + [operator]) * numOps
        specialized = specializeProgram(
            program,
            getNumRegions=getNumRegions,
            maxstack=maxStack,
            generalizeFirst=False,
        )
        stack_use = program_stack_use(specialized, getNumRegions=getNumRegions)
        assert maxStack - numRegions < stack_use < maxStack

    def test_maxstack_commands(self):
        # See if two commands with deep blends are merged into one
        numRegions = 400
        numOps = 2
        getNumRegions = lambda iv: numRegions
        blend_one = [i for i in range(1 + numRegions)] + [1, "blend"]
        operands = blend_one * 6
        operator = "rrcurveto"
        program = (operands + [operator]) * numOps
        specialized = specializeProgram(
            program,
            getNumRegions=getNumRegions,
            maxstack=maxStack,
            generalizeFirst=False,
        )
        assert specialized.index("rrcurveto") == len(specialized) - 1


class CFF2VFTestSpecialize(DataFilesHandler):
    def test_blend_round_trip(self):
        ttx_path = self.getpath("TestSparseCFF2VF.ttx")
        ttf_font = TTFont(recalcBBoxes=False, recalcTimestamp=False)
        ttf_font.importXML(ttx_path)
        fontGlyphList = ttf_font.getGlyphOrder()
        topDict = ttf_font["CFF2"].cff.topDictIndex[0]
        charstrings = topDict.CharStrings
        for glyphName in fontGlyphList:
            cs = charstrings[glyphName]
            cs.decompile()
            cmds = programToCommands(cs.program, getNumRegions=cs.getNumRegions)
            cmds_g = generalizeCommands(cmds)
            cmds = specializeCommands(cmds_g, generalizeFirst=False)
            program = commandsToProgram(cmds)
            assert program == cs.program, (program, cs.program)
            program = specializeProgram(program, getNumRegions=cs.getNumRegions)
            assert program == cs.program, (program, cs.program)
            program_g = generalizeProgram(program, getNumRegions=cs.getNumRegions)
            program = commandsToProgram(cmds_g)
            assert program == program_g, (program, program_g)

    def test_blend_programToCommands(self):
        ttx_path = self.getpath("TestCFF2Widths.ttx")
        ttf_font = TTFont(recalcBBoxes=False, recalcTimestamp=False)
        ttf_font.importXML(ttx_path)
        fontGlyphList = ttf_font.getGlyphOrder()
        topDict = ttf_font["CFF2"].cff.topDictIndex[0]
        charstrings = topDict.CharStrings
        for glyphName in fontGlyphList:
            cs = charstrings[glyphName]
            cs.decompile()
            cmds = programToCommands(cs.program, getNumRegions=cs.getNumRegions)
            program = commandsToProgram(cmds)
            assert program == cs.program, (program, cs.program)


if __name__ == "__main__":
    import sys

    sys.exit(pytest.main(sys.argv))



from fontTools.pens.recordingPen import (
    RecordingPen,
    DecomposingRecordingPen,
    RecordingPointPen,
)
import pytest


class _TestGlyph(object):
    def draw(self, pen):
        pen.moveTo((0.0, 0.0))
        pen.lineTo((0.0, 100.0))
        pen.curveTo((50.0, 75.0), (60.0, 50.0), (50.0, 0.0))
        pen.closePath()

    def drawPoints(self, pen):
        pen.beginPath(identifier="abc")
        pen.addPoint((0.0, 0.0), "line", False, "start", identifier="0000")
        pen.addPoint((0.0, 100.0), "line", False, None, identifier="0001")
        pen.addPoint((50.0, 75.0), None, False, None, identifier="0002")
        pen.addPoint((60.0, 50.0), None, False, None, identifier="0003")
        pen.addPoint((50.0, 0.0), "curve", True, "last", identifier="0004")
        pen.endPath()


class RecordingPenTest(object):
    def test_addComponent(self):
        pen = RecordingPen()
        pen.addComponent("a", (2, 0, 0, 3, -10, 5))
        assert pen.value == [("addComponent", ("a", (2, 0, 0, 3, -10, 5)))]


class DecomposingRecordingPenTest(object):
    def test_addComponent_decomposed(self):
        pen = DecomposingRecordingPen({"a": _TestGlyph()})
        pen.addComponent("a", (2, 0, 0, 3, -10, 5))
        assert pen.value == [
            ("moveTo", ((-10.0, 5.0),)),
            ("lineTo", ((-10.0, 305.0),)),
            ("curveTo", ((90.0, 230.0), (110.0, 155.0), (90.0, 5.0))),
            ("closePath", ()),
        ]

    def test_addComponent_missing_raises(self):
        pen = DecomposingRecordingPen(dict())
        with pytest.raises(KeyError) as excinfo:
            pen.addComponent("a", (1, 0, 0, 1, 0, 0))
        assert excinfo.value.args[0] == "a"


class RecordingPointPenTest:
    def test_record_and_replay(self):
        pen = RecordingPointPen()
        glyph = _TestGlyph()
        glyph.drawPoints(pen)
        pen.addComponent("a", (2, 0, 0, 2, -10, 5))

        assert pen.value == [
            ("beginPath", (), {"identifier": "abc"}),
            ("addPoint", ((0.0, 0.0), "line", False, "start"), {"identifier": "0000"}),
            ("addPoint", ((0.0, 100.0), "line", False, None), {"identifier": "0001"}),
            ("addPoint", ((50.0, 75.0), None, False, None), {"identifier": "0002"}),
            ("addPoint", ((60.0, 50.0), None, False, None), {"identifier": "0003"}),
            ("addPoint", ((50.0, 0.0), "curve", True, "last"), {"identifier": "0004"}),
            ("endPath", (), {}),
            ("addComponent", ("a", (2, 0, 0, 2, -10, 5)), {}),
        ]

        pen2 = RecordingPointPen()
        pen.replay(pen2)

        assert pen2.value == pen.value



import io
import fontTools.ttLib.tables.otBase
from fontTools.misc.testTools import getXML, stripVariableItemsFromTTX
from fontTools.misc.textTools import tobytes, tostr
from fontTools import subset
from fontTools.fontBuilder import FontBuilder
from fontTools.pens.ttGlyphPen import TTGlyphPen
from fontTools.ttLib import TTFont, newTable
from fontTools.ttLib.tables import otTables as ot
from fontTools.misc.loggingTools import CapturingLogHandler
from fontTools.subset.svg import etree
import difflib
import logging
import os
import shutil
import sys
import tempfile
import unittest
import pathlib
import pytest


class SubsetTest:
    @classmethod
    def setup_class(cls):
        cls.tempdir = None
        cls.num_tempfiles = 0

    @classmethod
    def teardown_class(cls):
        if cls.tempdir:
            shutil.rmtree(cls.tempdir, ignore_errors=True)

    @staticmethod
    def getpath(*testfile):
        path, _ = os.path.split(__file__)
        return os.path.join(path, "data", *testfile)

    @classmethod
    def temp_path(cls, suffix):
        if not cls.tempdir:
            cls.tempdir = tempfile.mkdtemp()
        cls.num_tempfiles += 1
        return os.path.join(cls.tempdir, "tmp%d%s" % (cls.num_tempfiles, suffix))

    @staticmethod
    def read_ttx(path):
        with open(path, "r", encoding="utf-8") as f:
            ttx = f.read()
        # don't care whether TTF or OTF, thus strip sfntVersion as well
        return stripVariableItemsFromTTX(ttx, sfntVersion=True).splitlines(True)

    def expect_ttx(self, font, expected_ttx, tables=None):
        path = self.temp_path(suffix=".ttx")
        font.saveXML(path, tables=tables)
        actual = self.read_ttx(path)
        expected = self.read_ttx(expected_ttx)
        if actual != expected:
            for line in difflib.unified_diff(
                expected, actual, fromfile=expected_ttx, tofile=path
            ):
                sys.stdout.write(line)
            pytest.fail("TTX output is different from expected")

    def compile_font(self, path, suffix):
        savepath = self.temp_path(suffix=suffix)
        font = TTFont(recalcBBoxes=False, recalcTimestamp=False)
        font.importXML(path)
        font.save(savepath, reorderTables=None)
        return savepath

    # -----
    # Tests
    # -----

    def test_layout_scripts(self):
        fontpath = self.compile_font(self.getpath("layout_scripts.ttx"), ".otf")
        subsetpath = self.temp_path(".otf")
        subset.main(
            [
                fontpath,
                "--glyphs=*",
                "--layout-features=*",
                "--layout-scripts=latn,arab.URD,arab.dflt",
                "--output-file=%s" % subsetpath,
            ]
        )
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(
            subsetfont, self.getpath("expect_layout_scripts.ttx"), ["GPOS", "GSUB"]
        )

    def test_no_notdef_outline_otf(self):
        fontpath = self.compile_font(self.getpath("TestOTF-Regular.ttx"), ".otf")
        subsetpath = self.temp_path(".otf")
        subset.main(
            [
                fontpath,
                "--no-notdef-outline",
                "--gids=0",
                "--output-file=%s" % subsetpath,
            ]
        )
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(
            subsetfont, self.getpath("expect_no_notdef_outline_otf.ttx"), ["CFF "]
        )

    def test_no_notdef_outline_cid(self):
        fontpath = self.compile_font(self.getpath("TestCID-Regular.ttx"), ".otf")
        subsetpath = self.temp_path(".otf")
        subset.main(
            [
                fontpath,
                "--no-notdef-outline",
                "--gids=0",
                "--output-file=%s" % subsetpath,
            ]
        )
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(
            subsetfont, self.getpath("expect_no_notdef_outline_cid.ttx"), ["CFF "]
        )

    def test_no_notdef_outline_ttf(self):
        fontpath = self.compile_font(self.getpath("TestTTF-Regular.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main(
            [
                fontpath,
                "--no-notdef-outline",
                "--gids=0",
                "--output-file=%s" % subsetpath,
            ]
        )
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(
            subsetfont,
            self.getpath("expect_no_notdef_outline_ttf.ttx"),
            ["glyf", "hmtx"],
        )

    def test_subset_ankr(self):
        fontpath = self.compile_font(self.getpath("TestANKR.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main([fontpath, "--glyphs=one", "--output-file=%s" % subsetpath])
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(subsetfont, self.getpath("expect_ankr.ttx"), ["ankr"])

    def test_subset_ankr_remove(self):
        fontpath = self.compile_font(self.getpath("TestANKR.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main([fontpath, "--glyphs=two", "--output-file=%s" % subsetpath])
        assert "ankr" not in TTFont(subsetpath)

    def test_subset_bsln_format_0(self):
        fontpath = self.compile_font(self.getpath("TestBSLN-0.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main([fontpath, "--glyphs=one", "--output-file=%s" % subsetpath])
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(subsetfont, self.getpath("expect_bsln_0.ttx"), ["bsln"])

    def test_subset_bsln_format_0_from_format_1(self):
        # TestBSLN-1 defines the ideographic baseline to be the font's default,
        # and specifies that glyphs {.notdef, zero, one, two} use the roman
        # baseline instead of the default ideographic baseline. As we request
        # a subsetted font with {zero, one} and the implicit .notdef, all
        # glyphs in the resulting font use the Roman baseline. In this case,
        # we expect a format 0 'bsln' table because it is the most compact.
        fontpath = self.compile_font(self.getpath("TestBSLN-1.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main(
            [fontpath, "--unicodes=U+0030-0031", "--output-file=%s" % subsetpath]
        )
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(subsetfont, self.getpath("expect_bsln_0.ttx"), ["bsln"])

    def test_subset_bsln_format_1(self):
        # TestBSLN-1 defines the ideographic baseline to be the font's default,
        # and specifies that glyphs {.notdef, zero, one, two} use the roman
        # baseline instead of the default ideographic baseline. We request
        # a subset where the majority of glyphs use the roman baseline,
        # but one single glyph (uni2EA2) is ideographic. In the resulting
        # subsetted font, we expect a format 1 'bsln' table whose default
        # is Roman, but with an override that uses the ideographic baseline
        # for uni2EA2.
        fontpath = self.compile_font(self.getpath("TestBSLN-1.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main(
            [fontpath, "--unicodes=U+0030-0031,U+2EA2", "--output-file=%s" % subsetpath]
        )
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(subsetfont, self.getpath("expect_bsln_1.ttx"), ["bsln"])

    def test_subset_bsln_format_2(self):
        # The 'bsln' table in TestBSLN-2 refers to control points in glyph 'P'
        # for defining its baselines. Therefore, the subsetted font should
        # include this glyph even though it is not requested explicitly.
        fontpath = self.compile_font(self.getpath("TestBSLN-2.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main([fontpath, "--glyphs=one", "--output-file=%s" % subsetpath])
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(subsetfont, self.getpath("expect_bsln_2.ttx"), ["bsln"])

    def test_subset_bsln_format_2_from_format_3(self):
        # TestBSLN-3 defines the ideographic baseline to be the font's default,
        # and specifies that glyphs {.notdef, zero, one, two, P} use the roman
        # baseline instead of the default ideographic baseline. As we request
        # a subsetted font with zero and the implicit .notdef and P for
        # baseline measurement, all glyphs in the resulting font use the Roman
        # baseline. In this case, we expect a format 2 'bsln' table because it
        # is the most compact encoding.
        fontpath = self.compile_font(self.getpath("TestBSLN-3.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main([fontpath, "--unicodes=U+0030", "--output-file=%s" % subsetpath])
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(subsetfont, self.getpath("expect_bsln_2.ttx"), ["bsln"])

    def test_subset_bsln_format_3(self):
        # TestBSLN-3 defines the ideographic baseline to be the font's default,
        # and specifies that glyphs {.notdef, zero, one, two} use the roman
        # baseline instead of the default ideographic baseline. We request
        # a subset where the majority of glyphs use the roman baseline,
        # but one single glyph (uni2EA2) is ideographic. In the resulting
        # subsetted font, we expect a format 1 'bsln' table whose default
        # is Roman, but with an override that uses the ideographic baseline
        # for uni2EA2.
        fontpath = self.compile_font(self.getpath("TestBSLN-3.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main(
            [fontpath, "--unicodes=U+0030-0031,U+2EA2", "--output-file=%s" % subsetpath]
        )
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(subsetfont, self.getpath("expect_bsln_3.ttx"), ["bsln"])

    def test_subset_clr(self):
        fontpath = self.compile_font(self.getpath("TestCLR-Regular.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main([fontpath, "--glyphs=smileface", "--output-file=%s" % subsetpath])
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(
            subsetfont,
            self.getpath("expect_keep_colr.ttx"),
            ["GlyphOrder", "hmtx", "glyf", "COLR", "CPAL"],
        )

    def test_subset_gvar(self):
        fontpath = self.compile_font(self.getpath("TestGVAR.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main(
            [fontpath, "--unicodes=U+002B,U+2212", "--output-file=%s" % subsetpath]
        )
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(
            subsetfont,
            self.getpath("expect_keep_gvar.ttx"),
            ["GlyphOrder", "avar", "fvar", "gvar", "name"],
        )

    def test_subset_gvar_notdef_outline(self):
        fontpath = self.compile_font(self.getpath("TestGVAR.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main(
            [
                fontpath,
                "--unicodes=U+0030",
                "--notdef_outline",
                "--output-file=%s" % subsetpath,
            ]
        )
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(
            subsetfont,
            self.getpath("expect_keep_gvar_notdef_outline.ttx"),
            ["GlyphOrder", "avar", "fvar", "gvar", "name"],
        )

    def test_subset_lcar_remove(self):
        fontpath = self.compile_font(self.getpath("TestLCAR-0.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main([fontpath, "--glyphs=one", "--output-file=%s" % subsetpath])
        subsetfont = TTFont(subsetpath)
        assert "lcar" not in subsetfont

    def test_subset_lcar_format_0(self):
        fontpath = self.compile_font(self.getpath("TestLCAR-0.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main([fontpath, "--unicodes=U+FB01", "--output-file=%s" % subsetpath])
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(subsetfont, self.getpath("expect_lcar_0.ttx"), ["lcar"])

    def test_subset_lcar_format_1(self):
        fontpath = self.compile_font(self.getpath("TestLCAR-1.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main([fontpath, "--unicodes=U+FB01", "--output-file=%s" % subsetpath])
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(subsetfont, self.getpath("expect_lcar_1.ttx"), ["lcar"])

    def test_subset_math(self):
        fontpath = self.compile_font(self.getpath("TestMATH-Regular.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main(
            [
                fontpath,
                "--unicodes=U+0041,U+0028,U+0302,U+1D400,U+1D435",
                "--output-file=%s" % subsetpath,
            ]
        )
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(
            subsetfont,
            self.getpath("expect_keep_math.ttx"),
            ["GlyphOrder", "CFF ", "MATH", "hmtx"],
        )

    def test_subset_math_partial(self):
        fontpath = self.compile_font(self.getpath("test_math_partial.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main([fontpath, "--text=A", "--output-file=%s" % subsetpath])
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(subsetfont, self.getpath("expect_math_partial.ttx"), ["MATH"])

    def test_subset_opbd_remove(self):
        # In the test font, only the glyphs 'A' and 'zero' have an entry in
        # the Optical Bounds table. When subsetting, we do not request any
        # of those glyphs. Therefore, the produced subsetted font should
        # not contain an 'opbd' table.
        fontpath = self.compile_font(self.getpath("TestOPBD-0.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main([fontpath, "--glyphs=one", "--output-file=%s" % subsetpath])
        subsetfont = TTFont(subsetpath)
        assert "opbd" not in subsetfont

    def test_subset_opbd_format_0(self):
        fontpath = self.compile_font(self.getpath("TestOPBD-0.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main([fontpath, "--glyphs=A", "--output-file=%s" % subsetpath])
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(subsetfont, self.getpath("expect_opbd_0.ttx"), ["opbd"])

    def test_subset_opbd_format_1(self):
        fontpath = self.compile_font(self.getpath("TestOPBD-1.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main([fontpath, "--glyphs=A", "--output-file=%s" % subsetpath])
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(subsetfont, self.getpath("expect_opbd_1.ttx"), ["opbd"])

    def test_subset_prop_remove_default_zero(self):
        # If all glyphs have an AAT glyph property with value 0,
        # the "prop" table should be removed from the subsetted font.
        fontpath = self.compile_font(self.getpath("TestPROP.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main([fontpath, "--unicodes=U+0041", "--output-file=%s" % subsetpath])
        subsetfont = TTFont(subsetpath)
        assert "prop" not in subsetfont

    def test_subset_prop_0(self):
        # If all glyphs share the same AAT glyph properties, the "prop" table
        # in the subsetted font should use format 0.
        #
        # Unless the shared value is zero, in which case the subsetted font
        # should have no "prop" table at all. But that case has already been
        # tested above in test_subset_prop_remove_default_zero().
        fontpath = self.compile_font(self.getpath("TestPROP.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main(
            [
                fontpath,
                "--unicodes=U+0030-0032",
                "--no-notdef-glyph",
                "--output-file=%s" % subsetpath,
            ]
        )
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(subsetfont, self.getpath("expect_prop_0.ttx"), ["prop"])

    def test_subset_prop_1(self):
        # If not all glyphs share the same AAT glyph properties, the subsetted
        # font should contain a "prop" table in format 1. To save space, the
        # DefaultProperties should be set to the most frequent value.
        fontpath = self.compile_font(self.getpath("TestPROP.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main(
            [
                fontpath,
                "--unicodes=U+0030-0032",
                "--notdef-outline",
                "--output-file=%s" % subsetpath,
            ]
        )
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(subsetfont, self.getpath("expect_prop_1.ttx"), ["prop"])

    def test_options(self):
        # https://github.com/fonttools/fonttools/issues/413
        opt1 = subset.Options()
        assert "Xyz-" not in opt1.layout_features
        opt2 = subset.Options()
        opt2.layout_features.append("Xyz-")
        assert "Xyz-" in opt2.layout_features
        assert "Xyz-" not in opt1.layout_features

    def test_google_color(self):
        fontpath = self.compile_font(self.getpath("google_color.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main([fontpath, "--gids=0,1", "--output-file=%s" % subsetpath])
        subsetfont = TTFont(subsetpath)
        assert "CBDT" in subsetfont
        assert "CBLC" in subsetfont
        assert "x" in subsetfont["CBDT"].strikeData[0]
        assert "y" not in subsetfont["CBDT"].strikeData[0]

    def test_google_color_all(self):
        fontpath = self.compile_font(self.getpath("google_color.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main([fontpath, "--unicodes=*", "--output-file=%s" % subsetpath])
        subsetfont = TTFont(subsetpath)
        assert "x" in subsetfont["CBDT"].strikeData[0]
        assert "y" in subsetfont["CBDT"].strikeData[0]

    def test_sbix(self):
        fontpath = self.compile_font(self.getpath("sbix.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main([fontpath, "--gids=0,1", "--output-file=%s" % subsetpath])
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(subsetfont, self.getpath("expect_sbix.ttx"), ["sbix"])

    def test_varComposite(self):
        fontpath = self.getpath("..", "..", "ttLib", "data", "varc-ac00-ac01.ttf")
        origfont = TTFont(fontpath)
        assert len(origfont.getGlyphOrder()) == 11
        subsetpath = self.temp_path(".ttf")
        subset.main([fontpath, "--unicodes=ac00", "--output-file=%s" % subsetpath])
        subsetfont = TTFont(subsetpath)
        assert len(subsetfont.getGlyphOrder()) == 6
        subset.main([fontpath, "--unicodes=ac01", "--output-file=%s" % subsetpath])
        subsetfont = TTFont(subsetpath)
        assert len(subsetfont.getGlyphOrder()) == 8

    def test_timing_publishes_parts(self):
        fontpath = self.compile_font(self.getpath("TestTTF-Regular.ttx"), ".ttf")

        options = subset.Options()
        options.timing = True
        subsetter = subset.Subsetter(options)
        subsetter.populate(text="ABC")
        font = TTFont(fontpath)
        with CapturingLogHandler("fontTools.subset.timer", logging.DEBUG) as captor:
            subsetter.subset(font)
        logs = captor.records

        assert len(logs) > 5
        assert len(logs) == len(
            [l for l in logs if "msg" in l.args and "time" in l.args]
        )
        # Look for a few things we know should happen
        assert filter(lambda l: l.args["msg"] == "load 'cmap'", logs)
        assert filter(lambda l: l.args["msg"] == "subset 'cmap'", logs)
        assert filter(lambda l: l.args["msg"] == "subset 'glyf'", logs)

    def test_passthrough_tables(self):
        fontpath = self.compile_font(self.getpath("TestTTF-Regular.ttx"), ".ttf")
        font = TTFont(fontpath)
        unknown_tag = "ZZZZ"
        unknown_table = newTable(unknown_tag)
        unknown_table.data = b"\0" * 10
        font[unknown_tag] = unknown_table
        font.save(fontpath)

        subsetpath = self.temp_path(".ttf")
        subset.main([fontpath, "--output-file=%s" % subsetpath])
        subsetfont = TTFont(subsetpath)

        # tables we can't subset are dropped by default
        assert unknown_tag not in subsetfont

        subsetpath = self.temp_path(".ttf")
        subset.main([fontpath, "--passthrough-tables", "--output-file=%s" % subsetpath])
        subsetfont = TTFont(subsetpath)

        # unknown tables are kept if --passthrough-tables option is passed
        assert unknown_tag in subsetfont

    def test_non_BMP_text_arg_input(self):
        fontpath = self.compile_font(
            self.getpath("TestTTF-Regular_non_BMP_char.ttx"), ".ttf"
        )
        subsetpath = self.temp_path(".ttf")
        text = tostr("A\U0001F6D2", encoding="utf-8")

        subset.main([fontpath, "--text=%s" % text, "--output-file=%s" % subsetpath])
        subsetfont = TTFont(subsetpath)

        assert subsetfont["maxp"].numGlyphs == 3
        assert subsetfont.getGlyphOrder() == [".notdef", "A", "u1F6D2"]

    def test_non_BMP_text_file_input(self):
        fontpath = self.compile_font(
            self.getpath("TestTTF-Regular_non_BMP_char.ttx"), ".ttf"
        )
        subsetpath = self.temp_path(".ttf")
        text = tobytes("A\U0001F6D2", encoding="utf-8")
        with tempfile.NamedTemporaryFile(delete=False) as tmp:
            tmp.write(text)

        try:
            subset.main(
                [fontpath, "--text-file=%s" % tmp.name, "--output-file=%s" % subsetpath]
            )
            subsetfont = TTFont(subsetpath)
        finally:
            os.remove(tmp.name)

        assert subsetfont["maxp"].numGlyphs == 3
        assert subsetfont.getGlyphOrder() == [".notdef", "A", "u1F6D2"]

    def test_no_hinting_CFF(self):
        ttxpath = self.getpath("Lobster.subset.ttx")
        fontpath = self.compile_font(ttxpath, ".otf")
        subsetpath = self.temp_path(".otf")
        subset.main(
            [
                fontpath,
                "--no-hinting",
                "--notdef-outline",
                "--output-file=%s" % subsetpath,
                "*",
            ]
        )
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(subsetfont, self.getpath("expect_no_hinting_CFF.ttx"), ["CFF "])

    def test_desubroutinize_CFF(self):
        ttxpath = self.getpath("Lobster.subset.ttx")
        fontpath = self.compile_font(ttxpath, ".otf")
        subsetpath = self.temp_path(".otf")
        subset.main(
            [
                fontpath,
                "--desubroutinize",
                "--notdef-outline",
                "--output-file=%s" % subsetpath,
                "*",
            ]
        )
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(
            subsetfont, self.getpath("expect_desubroutinize_CFF.ttx"), ["CFF "]
        )

    def test_desubroutinize_hinted_subrs_CFF(self):
        ttxpath = self.getpath("test_hinted_subrs_CFF.ttx")
        fontpath = self.compile_font(ttxpath, ".otf")
        subsetpath = self.temp_path(".otf")
        subset.main(
            [
                fontpath,
                "--desubroutinize",
                "--notdef-outline",
                "--output-file=%s" % subsetpath,
                "*",
            ]
        )
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(
            subsetfont, self.getpath("test_hinted_subrs_CFF.desub.ttx"), ["CFF "]
        )

    def test_desubroutinize_cntrmask_CFF(self):
        ttxpath = self.getpath("test_cntrmask_CFF.ttx")
        fontpath = self.compile_font(ttxpath, ".otf")
        subsetpath = self.temp_path(".otf")
        subset.main(
            [
                fontpath,
                "--desubroutinize",
                "--notdef-outline",
                "--output-file=%s" % subsetpath,
                "*",
            ]
        )
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(
            subsetfont, self.getpath("test_cntrmask_CFF.desub.ttx"), ["CFF "]
        )

    def test_no_hinting_desubroutinize_CFF(self):
        ttxpath = self.getpath("test_hinted_subrs_CFF.ttx")
        fontpath = self.compile_font(ttxpath, ".otf")
        subsetpath = self.temp_path(".otf")
        subset.main(
            [
                fontpath,
                "--no-hinting",
                "--desubroutinize",
                "--notdef-outline",
                "--output-file=%s" % subsetpath,
                "*",
            ]
        )
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(
            subsetfont,
            self.getpath("expect_no_hinting_desubroutinize_CFF.ttx"),
            ["CFF "],
        )

    def test_no_hinting_TTF(self):
        fontpath = self.compile_font(self.getpath("TestTTF-Regular.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main(
            [
                fontpath,
                "--no-hinting",
                "--notdef-outline",
                "--output-file=%s" % subsetpath,
                "*",
            ]
        )
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(
            subsetfont, self.getpath("expect_no_hinting_TTF.ttx"), ["glyf", "maxp"]
        )
        for tag in subset.Options().hinting_tables:
            assert tag not in subsetfont

    def test_notdef_width_cid(self):
        # https://github.com/fonttools/fonttools/pull/845
        fontpath = self.compile_font(self.getpath("NotdefWidthCID-Regular.ttx"), ".otf")
        subsetpath = self.temp_path(".otf")
        subset.main(
            [
                fontpath,
                "--no-notdef-outline",
                "--gids=0,1",
                "--output-file=%s" % subsetpath,
            ]
        )
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(
            subsetfont, self.getpath("expect_notdef_width_cid.ttx"), ["CFF "]
        )

    def test_recalc_bounds_ttf(self):
        ttxpath = self.getpath("TestTTF-Regular.ttx")
        font = TTFont()
        font.importXML(ttxpath)
        head = font["head"]
        bounds = [head.xMin, head.yMin, head.xMax, head.yMax]

        fontpath = self.compile_font(ttxpath, ".ttf")
        subsetpath = self.temp_path(".ttf")

        # by default, the subsetter does not recalculate the bounding box
        subset.main([fontpath, "--output-file=%s" % subsetpath, "*"])
        head = TTFont(subsetpath)["head"]
        assert bounds == [head.xMin, head.yMin, head.xMax, head.yMax]

        subset.main([fontpath, "--recalc-bounds", "--output-file=%s" % subsetpath, "*"])
        head = TTFont(subsetpath)["head"]
        bounds = [132, 304, 365, 567]
        assert bounds == [head.xMin, head.yMin, head.xMax, head.yMax]

    def test_recalc_bounds_otf(self):
        ttxpath = self.getpath("TestOTF-Regular.ttx")
        font = TTFont()
        font.importXML(ttxpath)
        head = font["head"]
        bounds = [head.xMin, head.yMin, head.xMax, head.yMax]

        fontpath = self.compile_font(ttxpath, ".otf")
        subsetpath = self.temp_path(".otf")

        # by default, the subsetter does not recalculate the bounding box
        subset.main([fontpath, "--output-file=%s" % subsetpath, "*"])
        head = TTFont(subsetpath)["head"]
        assert bounds == [head.xMin, head.yMin, head.xMax, head.yMax]

        subset.main([fontpath, "--recalc-bounds", "--output-file=%s" % subsetpath, "*"])
        head = TTFont(subsetpath)["head"]
        bounds = [132, 304, 365, 567]
        assert bounds == [head.xMin, head.yMin, head.xMax, head.yMax]

    def test_recalc_timestamp_ttf(self):
        ttxpath = self.getpath("TestTTF-Regular.ttx")
        font = TTFont()
        font.importXML(ttxpath)
        modified = font["head"].modified
        fontpath = self.compile_font(ttxpath, ".ttf")
        subsetpath = self.temp_path(".ttf")

        # by default, the subsetter does not recalculate the modified timestamp
        subset.main([fontpath, "--output-file=%s" % subsetpath, "*"])
        assert modified == TTFont(subsetpath)["head"].modified

        subset.main(
            [fontpath, "--recalc-timestamp", "--output-file=%s" % subsetpath, "*"]
        )
        assert modified < TTFont(subsetpath)["head"].modified

    def test_recalc_timestamp_otf(self):
        ttxpath = self.getpath("TestOTF-Regular.ttx")
        font = TTFont()
        font.importXML(ttxpath)
        modified = font["head"].modified
        fontpath = self.compile_font(ttxpath, ".otf")
        subsetpath = self.temp_path(".otf")

        # by default, the subsetter does not recalculate the modified timestamp
        subset.main([fontpath, "--output-file=%s" % subsetpath, "*"])
        assert modified == TTFont(subsetpath)["head"].modified

        subset.main(
            [fontpath, "--recalc-timestamp", "--output-file=%s" % subsetpath, "*"]
        )
        assert modified < TTFont(subsetpath)["head"].modified

    def test_recalc_max_context(self):
        ttxpath = self.getpath("Lobster.subset.ttx")
        font = TTFont()
        font.importXML(ttxpath)
        max_context = font["OS/2"].usMaxContext
        fontpath = self.compile_font(ttxpath, ".otf")
        subsetpath = self.temp_path(".otf")

        # by default, the subsetter does not recalculate the usMaxContext
        subset.main(
            [fontpath, "--drop-tables+=GSUB,GPOS", "--output-file=%s" % subsetpath]
        )
        assert max_context == TTFont(subsetpath)["OS/2"].usMaxContext

        subset.main(
            [
                fontpath,
                "--recalc-max-context",
                "--drop-tables+=GSUB,GPOS",
                "--output-file=%s" % subsetpath,
            ]
        )
        assert 0 == TTFont(subsetpath)["OS/2"].usMaxContext

    def test_retain_gids_ttf(self):
        fontpath = self.compile_font(self.getpath("TestTTF-Regular.ttx"), ".ttf")
        font = TTFont(fontpath)

        assert font["hmtx"]["A"] == (500, 132)
        assert font["hmtx"]["B"] == (400, 132)

        assert font["glyf"]["A"].numberOfContours > 0
        assert font["glyf"]["B"].numberOfContours > 0

        subsetpath = self.temp_path(".ttf")
        subset.main(
            [
                fontpath,
                "--retain-gids",
                "--output-file=%s" % subsetpath,
                "--glyph-names",
                "B",
            ]
        )
        subsetfont = TTFont(subsetpath)

        assert subsetfont.getGlyphOrder() == font.getGlyphOrder()[0:3]

        hmtx = subsetfont["hmtx"]
        assert hmtx["A"] == (0, 0)
        assert hmtx["B"] == (400, 132)

        glyf = subsetfont["glyf"]
        assert glyf["A"].numberOfContours == 0
        assert glyf["B"].numberOfContours > 0

    def test_retain_gids_cff(self):
        fontpath = self.compile_font(self.getpath("TestOTF-Regular.ttx"), ".otf")
        font = TTFont(fontpath)

        assert font["hmtx"]["A"] == (500, 132)
        assert font["hmtx"]["B"] == (400, 132)
        assert font["hmtx"]["C"] == (500, 0)

        font["CFF "].cff[0].decompileAllCharStrings()
        cs = font["CFF "].cff[0].CharStrings
        assert len(cs["A"].program) > 0
        assert len(cs["B"].program) > 0
        assert len(cs["C"].program) > 0

        subsetpath = self.temp_path(".otf")
        subset.main(
            [
                fontpath,
                "--retain-gids",
                "--output-file=%s" % subsetpath,
                "--glyph-names",
                "B",
            ]
        )
        subsetfont = TTFont(subsetpath)

        assert subsetfont.getGlyphOrder() == font.getGlyphOrder()[0:3]

        hmtx = subsetfont["hmtx"]
        assert hmtx["A"] == (0, 0)
        assert hmtx["B"] == (400, 132)

        subsetfont["CFF "].cff[0].decompileAllCharStrings()
        cs = subsetfont["CFF "].cff[0].CharStrings

        assert cs["A"].program == ["endchar"]
        assert len(cs["B"].program) > 0

    def test_retain_gids_cff2(self):
        ttx_path = self.getpath(
            "../../varLib/data/master_ttx_varfont_otf/TestCFF2VF.ttx"
        )
        fontpath = self.compile_font(ttx_path, ".otf")
        font = TTFont(fontpath)

        assert font["hmtx"]["A"] == (600, 31)
        assert font["hmtx"]["T"] == (600, 41)

        font["CFF2"].cff[0].decompileAllCharStrings()
        cs = font["CFF2"].cff[0].CharStrings
        assert len(cs["A"].program) > 0
        assert len(cs["T"].program) > 0

        subsetpath = self.temp_path(".otf")
        subset.main(
            [
                fontpath,
                "--retain-gids",
                "--output-file=%s" % subsetpath,
                "T",
            ]
        )
        subsetfont = TTFont(subsetpath)

        assert len(subsetfont.getGlyphOrder()) == len(font.getGlyphOrder()[0:3])

        hmtx = subsetfont["hmtx"]
        assert hmtx["glyph00001"] == (0, 0)
        assert hmtx["T"] == (600, 41)

        subsetfont["CFF2"].cff[0].decompileAllCharStrings()
        cs = subsetfont["CFF2"].cff[0].CharStrings
        assert cs["glyph00001"].program == []
        assert len(cs["T"].program) > 0

    def test_HVAR_VVAR(self):
        fontpath = self.compile_font(self.getpath("TestHVVAR.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main([fontpath, "--text=BD", "--output-file=%s" % subsetpath])
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(
            subsetfont,
            self.getpath("expect_HVVAR.ttx"),
            ["GlyphOrder", "HVAR", "VVAR", "avar", "fvar"],
        )

    def test_HVAR_VVAR_retain_gids(self):
        fontpath = self.compile_font(self.getpath("TestHVVAR.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main(
            [fontpath, "--text=BD", "--retain-gids", "--output-file=%s" % subsetpath]
        )
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(
            subsetfont,
            self.getpath("expect_HVVAR_retain_gids.ttx"),
            ["GlyphOrder", "HVAR", "VVAR", "avar", "fvar"],
        )

    def test_subset_flavor_woff(self):
        fontpath = self.compile_font(self.getpath("TestTTF-Regular.ttx"), ".ttf")
        woff_path = self.temp_path(".woff")

        subset.main(
            [
                fontpath,
                "*",
                "--flavor=woff",
                "--output-file=%s" % woff_path,
            ]
        )
        woff = TTFont(woff_path)

        assert woff.flavor == "woff"

    def test_subset_flavor_woff2(self):
        # skip if brotli is not importable, required for woff2
        pytest.importorskip("brotli")

        fontpath = self.compile_font(self.getpath("TestTTF-Regular.ttx"), ".ttf")
        woff2_path = self.temp_path(".woff2")

        subset.main(
            [
                fontpath,
                "*",
                "--flavor=woff2",
                "--output-file=%s" % woff2_path,
            ]
        )
        woff2 = TTFont(woff2_path)

        assert woff2.flavor == "woff2"

    def test_subset_flavor_none(self):
        fontpath = self.compile_font(self.getpath("TestTTF-Regular.ttx"), ".ttf")
        ttf_path = self.temp_path(".ttf")

        subset.main(
            [
                fontpath,
                "*",
                "--output-file=%s" % ttf_path,
            ]
        )
        ttf = TTFont(ttf_path)

        assert ttf.flavor is None

    def test_subset_context_subst_format_3(self):
        # https://github.com/fonttools/fonttools/issues/1879
        # Test font contains 'calt' feature with Format 3 ContextSubst lookup subtables
        ttx = self.getpath("TestContextSubstFormat3.ttx")
        fontpath = self.compile_font(ttx, ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main([fontpath, "--unicodes=*", "--output-file=%s" % subsetpath])
        subsetfont = TTFont(subsetpath)
        # check all glyphs are kept via GSUB closure, no changes expected
        self.expect_ttx(subsetfont, ttx)

    def test_cmap_prune_format12(self):
        fontpath = self.compile_font(self.getpath("CmapSubsetTest.ttx"), ".ttf")
        subsetpath = self.temp_path(".ttf")
        subset.main([fontpath, "--glyphs=a", "--output-file=%s" % subsetpath])
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(subsetfont, self.getpath("CmapSubsetTest.subset.ttx"), ["cmap"])

    def test_cmap_format14(self):
        fontpath = self.compile_font(self.getpath("cmap14_font1.ttx"), ".otf")
        subsetpath = self.temp_path(".otf")

        subset.main([fontpath, "--unicodes=4e05", "--output-file=%s" % subsetpath])
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(subsetfont, self.getpath("cmap14_font1.no_uvs.ttx"), ["cmap"])

        subset.main(
            [fontpath, "--unicodes=4e05,e0100", "--output-file=%s" % subsetpath]
        )
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(subsetfont, self.getpath("cmap14_font1.uvs.ttx"), ["cmap"])

        subset.main([fontpath, "--unicodes=4e10", "--output-file=%s" % subsetpath])
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(
            subsetfont, self.getpath("cmap14_font1.no_uvs_non_default.ttx"), ["cmap"]
        )

        subset.main(
            [fontpath, "--unicodes=4e10,e0100", "--output-file=%s" % subsetpath]
        )
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(
            subsetfont, self.getpath("cmap14_font1.uvs_non_default.ttx"), ["cmap"]
        )

    @pytest.mark.parametrize("text, n", [("!", 1), ("#", 2)])
    def test_GPOS_PairPos_Format2_useClass0(self, text, n):
        # Check two things related to class 0 ('every other glyph'):
        # 1) that it's reused for ClassDef1 when it becomes empty as the subset glyphset
        #    is intersected with the table's Coverage
        # 2) that it is never reused for ClassDef2 even when it happens to become empty
        #    because of the subset glyphset. In this case, we don't keep a PairPosClass2
        #    subtable if only ClassDef2's class0 survived subsetting.
        # The test font (from Harfbuzz test suite) is constructed to trigger these two
        # situations depending on the input subset --text.
        # https://github.com/fonttools/fonttools/pull/2221
        fontpath = self.compile_font(
            self.getpath("GPOS_PairPos_Format2_PR_2221.ttx"), ".ttf"
        )
        subsetpath = self.temp_path(".ttf")

        expected_ttx = self.getpath(
            f"GPOS_PairPos_Format2_ClassDef{n}_useClass0.subset.ttx"
        )
        subset.main(
            [
                fontpath,
                f"--text='{text}'",
                "--layout-features+=test",
                "--output-file=%s" % subsetpath,
            ]
        )
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(subsetfont, expected_ttx, ["GPOS"])

    def test_GPOS_SinglePos_prune_post_subset_no_value(self):
        fontpath = self.compile_font(
            self.getpath("GPOS_SinglePos_no_value_issue_2312.ttx"), ".ttf"
        )
        subsetpath = self.temp_path(".ttf")
        subset.main([fontpath, "*", "--glyph-names", "--output-file=%s" % subsetpath])
        subsetfont = TTFont(subsetpath)
        self.expect_ttx(
            subsetfont,
            self.getpath("GPOS_SinglePos_no_value_issue_2312.subset.ttx"),
            ["GlyphOrder", "GPOS"],
        )

    @pytest.mark.parametrize(
        "installed, enabled, ok",
        [
            pytest.param(True, None, True, id="installed-auto-ok"),
            pytest.param(True, None, False, id="installed-auto-fail"),
            pytest.param(True, True, True, id="installed-enabled-ok"),
            pytest.param(True, True, False, id="installed-enabled-fail"),
            pytest.param(True, False, True, id="installed-disabled"),
            pytest.param(False, True, True, id="not_installed-enabled"),
            pytest.param(False, False, True, id="not_installed-disabled"),
        ],
    )
    def test_harfbuzz_repacker(self, caplog, monkeypatch, installed, enabled, ok):
        # Use a mock to test the pure-python serializer is used when uharfbuzz
        # returns an error or is not installed
        have_uharfbuzz = fontTools.ttLib.tables.otBase.have_uharfbuzz
        if installed:
            if not have_uharfbuzz:
                pytest.skip("uharfbuzz is not installed")
            if not ok:
                # pretend hb.repack/repack_with_tag return an error
                import uharfbuzz as hb

                def mock_repack(data, obj_list):
                    raise hb.RepackerError("mocking")

                monkeypatch.setattr(hb, "repack", mock_repack)

                if hasattr(hb, "repack_with_tag"):  # uharfbuzz >= 0.30.0

                    def mock_repack_with_tag(tag, data, obj_list):
                        raise hb.RepackerError("mocking")

                    monkeypatch.setattr(hb, "repack_with_tag", mock_repack_with_tag)
        else:
            if have_uharfbuzz:
                # pretend uharfbuzz is not installed
                monkeypatch.setattr(
                    fontTools.ttLib.tables.otBase, "have_uharfbuzz", False
                )

        fontpath = self.compile_font(self.getpath("harfbuzz_repacker.ttx"), ".otf")
        subsetpath = self.temp_path(".otf")
        args = [
            fontpath,
            "--unicodes=0x53a9",
            "--layout-features=*",
            f"--output-file={subsetpath}",
        ]
        if enabled is True:
            args.append("--harfbuzz-repacker")
        elif enabled is False:
            args.append("--no-harfbuzz-repacker")
        # elif enabled is None: ... is the default

        if enabled is True and not installed:
            # raise if enabled but not installed
            with pytest.raises(ImportError, match="uharfbuzz"):
                subset.main(args)
            return

        with caplog.at_level(logging.DEBUG, "fontTools.ttLib.tables.otBase"):
            subset.main(args)

        subsetfont = TTFont(subsetpath)
        # both hb.repack and pure-python serializer compile to the same ttx
        self.expect_ttx(
            subsetfont, self.getpath("expect_harfbuzz_repacker.ttx"), ["GSUB"]
        )

        if enabled or enabled is None:
            if installed:
                assert "serializing 'GSUB' with hb.repack" in caplog.text

        if enabled is None and not installed:
            assert (
                "uharfbuzz not found, compiling 'GSUB' with pure-python serializer"
            ) in caplog.text

        if enabled is False:
            assert (
                "hb.repack disabled, compiling 'GSUB' with pure-python serializer"
            ) in caplog.text

        # test we emit a log.error if hb.repack fails (and we don't if successful)
        assert (
            (
                "hb.repack failed to serialize 'GSUB', attempting fonttools resolutions "
                "; the error message was: RepackerError: mocking"
            )
            in caplog.text
        ) ^ ok

    def test_retain_east_asian_spacing_features(self):
        # This test font contains halt and vhal features, check that
        # they are retained by default after subsetting.
        ttx_path = self.getpath("NotoSansCJKjp-Regular.subset.ttx")
        ttx = pathlib.Path(ttx_path).read_text()
        assert 'FeatureTag value="halt"' in ttx
        assert 'FeatureTag value="vhal"' in ttx

        fontpath = self.compile_font(ttx_path, ".otf")
        subsetpath = self.temp_path(".otf")
        subset.main(
            [
                fontpath,
                "--unicodes=*",
                "--output-file=%s" % subsetpath,
            ]
        )
        # subset output is the same as the input
        self.expect_ttx(TTFont(subsetpath), ttx_path)


@pytest.fixture
def featureVarsTestFont():
    fb = FontBuilder(unitsPerEm=100)
    fb.setupGlyphOrder([".notdef", "f", "f_f", "dollar", "dollar.rvrn"])
    fb.setupCharacterMap({ord("f"): "f", ord("$"): "dollar"})
    fb.setupNameTable({"familyName": "TestFeatureVars", "styleName": "Regular"})
    fb.setupPost()
    fb.setupFvar(axes=[("wght", 100, 400, 900, "Weight")], instances=[])
    fb.addOpenTypeFeatures(
        """\
        feature dlig {
            sub f f by f_f;
        } dlig;
    """
    )
    fb.addFeatureVariations(
        [([{"wght": (0.20886, 1.0)}], {"dollar": "dollar.rvrn"})], featureTag="rvrn"
    )
    buf = io.BytesIO()
    fb.save(buf)
    buf.seek(0)

    return TTFont(buf)


def test_subset_feature_variations_keep_all(featureVarsTestFont):
    font = featureVarsTestFont

    options = subset.Options()
    subsetter = subset.Subsetter(options)
    subsetter.populate(unicodes=[ord("f"), ord("$")])
    subsetter.subset(font)

    featureTags = {r.FeatureTag for r in font["GSUB"].table.FeatureList.FeatureRecord}
    # 'dlig' is discretionary so it is dropped by default
    assert "dlig" not in featureTags
    assert "f_f" not in font.getGlyphOrder()
    # 'rvrn' is required so it is kept by default
    assert "rvrn" in featureTags
    assert "dollar.rvrn" in font.getGlyphOrder()


def test_subset_feature_variations_drop_all(featureVarsTestFont):
    font = featureVarsTestFont

    options = subset.Options()
    options.layout_features.remove("rvrn")  # drop 'rvrn'
    subsetter = subset.Subsetter(options)
    subsetter.populate(unicodes=[ord("f"), ord("$")])
    subsetter.subset(font)

    featureTags = {r.FeatureTag for r in font["GSUB"].table.FeatureList.FeatureRecord}
    glyphs = set(font.getGlyphOrder())

    assert "rvrn" not in featureTags
    assert glyphs == {".notdef", "f", "dollar"}
    # all FeatureVariationRecords were dropped
    assert font["GSUB"].table.FeatureVariations is None
    assert font["GSUB"].table.Version == 0x00010000


# TODO test_subset_feature_variations_drop_from_end_empty_records
# https://github.com/fonttools/fonttools/issues/1881#issuecomment-619415044


@pytest.fixture
def singlepos2_font():
    fb = FontBuilder(unitsPerEm=1000)
    fb.setupGlyphOrder([".notdef", "a", "b", "c"])
    fb.setupCharacterMap({ord("a"): "a", ord("b"): "b", ord("c"): "c"})
    fb.setupNameTable({"familyName": "TestSingePosFormat", "styleName": "Regular"})
    fb.setupPost()
    fb.addOpenTypeFeatures(
        """
        feature kern {
            pos a -50;
            pos b -40;
            pos c -50;
        } kern;
    """
    )

    buf = io.BytesIO()
    fb.save(buf)
    buf.seek(0)

    return TTFont(buf)


def test_subset_single_pos_format(singlepos2_font):
    font = singlepos2_font
    # The input font has a SinglePos Format 2 subtable where each glyph has
    # different ValueRecords
    assert getXML(font["GPOS"].table.LookupList.Lookup[0].toXML, font) == [
        "<Lookup>",
        '  <LookupType value="1"/>',
        '  <LookupFlag value="0"/>',
        "  <!-- SubTableCount=1 -->",
        '  <SinglePos index="0" Format="2">',
        "    <Coverage>",
        '      <Glyph value="a"/>',
        '      <Glyph value="b"/>',
        '      <Glyph value="c"/>',
        "    </Coverage>",
        '    <ValueFormat value="4"/>',
        "    <!-- ValueCount=3 -->",
        '    <Value index="0" XAdvance="-50"/>',
        '    <Value index="1" XAdvance="-40"/>',
        '    <Value index="2" XAdvance="-50"/>',
        "  </SinglePos>",
        "</Lookup>",
    ]

    options = subset.Options()
    subsetter = subset.Subsetter(options)
    subsetter.populate(unicodes=[ord("a"), ord("c")])
    subsetter.subset(font)

    # All the subsetted glyphs from the original SinglePos Format2 subtable
    # now have the same ValueRecord, so we use a more compact Format 1 subtable.
    assert getXML(font["GPOS"].table.LookupList.Lookup[0].toXML, font) == [
        "<Lookup>",
        '  <LookupType value="1"/>',
        '  <LookupFlag value="0"/>',
        "  <!-- SubTableCount=1 -->",
        '  <SinglePos index="0" Format="1">',
        "    <Coverage>",
        '      <Glyph value="a"/>',
        '      <Glyph value="c"/>',
        "    </Coverage>",
        '    <ValueFormat value="4"/>',
        '    <Value XAdvance="-50"/>',
        "  </SinglePos>",
        "</Lookup>",
    ]


def test_subset_single_pos_format2_all_None(singlepos2_font):
    # https://github.com/fonttools/fonttools/issues/2602
    font = singlepos2_font
    gpos = font["GPOS"].table
    subtable = gpos.LookupList.Lookup[0].SubTable[0]
    assert subtable.Format == 2
    # Hack a SinglePosFormat2 with ValueFormat = 0; our own buildSinglePos
    # never makes these as a SinglePosFormat1 is more compact, but they can
    # be found in the wild.
    subtable.Value = [None] * subtable.ValueCount
    subtable.ValueFormat = 0

    assert getXML(subtable.toXML, font) == [
        '<SinglePos Format="2">',
        "  <Coverage>",
        '    <Glyph value="a"/>',
        '    <Glyph value="b"/>',
        '    <Glyph value="c"/>',
        "  </Coverage>",
        '  <ValueFormat value="0"/>',
        "  <!-- ValueCount=3 -->",
        "</SinglePos>",
    ]

    options = subset.Options()
    subsetter = subset.Subsetter(options)
    subsetter.populate(unicodes=[ord("a"), ord("c")])
    subsetter.subset(font)

    # Check it was downgraded to Format1 after subsetting
    assert getXML(font["GPOS"].table.LookupList.Lookup[0].SubTable[0].toXML, font) == [
        '<SinglePos Format="1">',
        "  <Coverage>",
        '    <Glyph value="a"/>',
        '    <Glyph value="c"/>',
        "  </Coverage>",
        '  <ValueFormat value="0"/>',
        "</SinglePos>",
    ]


@pytest.fixture
def ttf_path(tmp_path):
    # $(dirname $0)/../ttLib/data
    ttLib_data = pathlib.Path(__file__).parent.parent / "ttLib" / "data"
    font = TTFont()
    font.importXML(ttLib_data / "TestTTF-Regular.ttx")
    font_path = tmp_path / "TestTTF-Regular.ttf"
    font.save(font_path)
    return font_path


def test_subset_empty_glyf(tmp_path, ttf_path):
    subset_path = tmp_path / (ttf_path.name + ".subset")
    # only keep empty .notdef and space glyph, resulting in an empty glyf table
    subset.main(
        [
            str(ttf_path),
            "--no-notdef-outline",
            "--glyph-names",
            f"--output-file={subset_path}",
            "--glyphs=.notdef space",
        ]
    )
    subset_font = TTFont(subset_path)

    assert subset_font.getGlyphOrder() == [".notdef", "space"]
    assert subset_font.reader["glyf"] == b"\x00"

    glyf = subset_font["glyf"]
    assert all(glyf[g].numberOfContours == 0 for g in subset_font.getGlyphOrder())

    loca = subset_font["loca"]
    assert all(loc == 0 for loc in loca)


@pytest.fixture
def colrv1_path(tmp_path):
    base_glyph_names = ["uni%04X" % i for i in range(0xE000, 0xE000 + 10)]
    layer_glyph_names = ["glyph%05d" % i for i in range(10, 20)]
    glyph_order = [".notdef"] + base_glyph_names + layer_glyph_names

    pen = TTGlyphPen(glyphSet=None)
    pen.moveTo((0, 0))
    pen.lineTo((0, 500))
    pen.lineTo((500, 500))
    pen.lineTo((500, 0))
    pen.closePath()
    glyph = pen.glyph()
    glyphs = {g: glyph for g in glyph_order}

    fb = FontBuilder(unitsPerEm=1024, isTTF=True)
    fb.setupGlyphOrder(glyph_order)
    fb.setupCharacterMap({int(name[3:], 16): name for name in base_glyph_names})
    fb.setupGlyf(glyphs)
    fb.setupHorizontalMetrics({g: (500, 0) for g in glyph_order})
    fb.setupHorizontalHeader()
    fb.setupOS2()
    fb.setupPost()
    fb.setupNameTable({"familyName": "TestCOLRv1", "styleName": "Regular"})

    fb.setupCOLR(
        {
            "uniE000": (
                ot.PaintFormat.PaintColrLayers,
                [
                    {
                        "Format": ot.PaintFormat.PaintGlyph,
                        "Paint": (ot.PaintFormat.PaintSolid, 0),
                        "Glyph": "glyph00010",
                    },
                    {
                        "Format": ot.PaintFormat.PaintGlyph,
                        "Paint": (ot.PaintFormat.PaintSolid, 2, 0.3),
                        "Glyph": "glyph00011",
                    },
                ],
            ),
            "uniE001": (
                ot.PaintFormat.PaintColrLayers,
                [
                    {
                        "Format": ot.PaintFormat.PaintTransform,
                        "Paint": {
                            "Format": ot.PaintFormat.PaintGlyph,
                            "Paint": {
                                "Format": ot.PaintFormat.PaintRadialGradient,
                                "x0": 250,
                                "y0": 250,
                                "r0": 250,
                                "x1": 200,
                                "y1": 200,
                                "r1": 0,
                                "ColorLine": {
                                    "ColorStop": [(0.0, 1), (1.0, 2)],
                                    "Extend": "repeat",
                                },
                            },
                            "Glyph": "glyph00012",
                        },
                        "Transform": (0.7071, 0.7071, -0.7071, 0.7071, 0, 0),
                    },
                    {
                        "Format": ot.PaintFormat.PaintGlyph,
                        "Paint": (ot.PaintFormat.PaintSolid, 1, 0.5),
                        "Glyph": "glyph00013",
                    },
                ],
            ),
            "uniE002": (
                ot.PaintFormat.PaintColrLayers,
                [
                    {
                        "Format": ot.PaintFormat.PaintGlyph,
                        "Paint": {
                            "Format": ot.PaintFormat.PaintLinearGradient,
                            "x0": 0,
                            "y0": 0,
                            "x1": 500,
                            "y1": 500,
                            "x2": -500,
                            "y2": 500,
                            "ColorLine": {"ColorStop": [(0.0, 1), (1.0, 2)]},
                        },
                        "Glyph": "glyph00014",
                    },
                    {
                        "Format": ot.PaintFormat.PaintTransform,
                        "Paint": {
                            "Format": ot.PaintFormat.PaintGlyph,
                            "Paint": (ot.PaintFormat.PaintSolid, 1),
                            "Glyph": "glyph00015",
                        },
                        "Transform": (1, 0, 0, 1, 400, 400),
                    },
                ],
            ),
            "uniE003": {
                "Format": ot.PaintFormat.PaintRotateAroundCenter,
                "Paint": {
                    "Format": ot.PaintFormat.PaintColrGlyph,
                    "Glyph": "uniE001",
                },
                "angle": 45,
                "centerX": 250,
                "centerY": 250,
            },
            "uniE004": [
                ("glyph00016", 1),
                ("glyph00017", 0xFFFF),  # special palette index for foreground text
                ("glyph00018", 2),
            ],
        },
        clipBoxes={
            "uniE000": (0, 0, 200, 300),
            "uniE001": (0, 0, 500, 500),
            "uniE002": (-50, -50, 400, 400),
            "uniE003": (-50, -50, 400, 400),
        },
    )
    fb.setupCPAL(
        [
            [
                (1.0, 0.0, 0.0, 1.0),  # red
                (0.0, 1.0, 0.0, 1.0),  # green
                (0.0, 0.0, 1.0, 1.0),  # blue
            ],
        ],
    )

    output_path = tmp_path / "TestCOLRv1.ttf"
    fb.save(output_path)

    return output_path


@pytest.fixture
def colrv1_cpalv1_path(colrv1_path):
    # upgrade CPAL from v0 to v1 by adding labels
    font = TTFont(colrv1_path)
    fb = FontBuilder(font=font)
    fb.setupCPAL(
        [
            [
                (1.0, 0.0, 0.0, 1.0),  # red
                (0.0, 1.0, 0.0, 1.0),  # green
                (0.0, 0.0, 1.0, 1.0),  # blue
            ],
        ],
        paletteLabels=["test palette"],
        paletteEntryLabels=["first color", "second color", "third color"],
    )

    output_path = colrv1_path.parent / "TestCOLRv1CPALv1.ttf"
    fb.save(output_path)

    return output_path


@pytest.fixture
def colrv1_cpalv1_share_nameID_path(colrv1_path):
    font = TTFont(colrv1_path)
    fb = FontBuilder(font=font)
    fb.setupCPAL(
        [
            [
                (1.0, 0.0, 0.0, 1.0),  # red
                (0.0, 1.0, 0.0, 1.0),  # green
                (0.0, 0.0, 1.0, 1.0),  # blue
            ],
        ],
        paletteLabels=["test palette"],
        paletteEntryLabels=["first color", "second color", "third color"],
    )

    # Set the name ID of the first color to use nameID 1 = familyName = "TestCOLRv1"
    fb.font["CPAL"].paletteEntryLabels[0] = 1

    output_path = colrv1_path.parent / "TestCOLRv1CPALv1.ttf"
    fb.save(output_path)

    return output_path


def test_subset_COLRv1_and_CPAL(colrv1_path):
    subset_path = colrv1_path.parent / (colrv1_path.name + ".subset")

    subset.main(
        [
            str(colrv1_path),
            "--glyph-names",
            f"--output-file={subset_path}",
            "--unicodes=E002,E003,E004",
        ]
    )
    subset_font = TTFont(subset_path)

    glyph_set = set(subset_font.getGlyphOrder())

    # uniE000 and its children are excluded from subset
    assert "uniE000" not in glyph_set
    assert "glyph00010" not in glyph_set
    assert "glyph00011" not in glyph_set

    # uniE001 and children are pulled in indirectly as PaintColrGlyph by uniE003
    assert "uniE001" in glyph_set
    assert "glyph00012" in glyph_set
    assert "glyph00013" in glyph_set

    assert "uniE002" in glyph_set
    assert "glyph00014" in glyph_set
    assert "glyph00015" in glyph_set

    assert "uniE003" in glyph_set

    assert "uniE004" in glyph_set
    assert "glyph00016" in glyph_set
    assert "glyph00017" in glyph_set
    assert "glyph00018" in glyph_set

    assert "COLR" in subset_font
    colr = subset_font["COLR"].table
    assert colr.Version == 1
    assert len(colr.BaseGlyphRecordArray.BaseGlyphRecord) == 1
    assert len(colr.BaseGlyphList.BaseGlyphPaintRecord) == 3  # was 4

    base = colr.BaseGlyphList.BaseGlyphPaintRecord[0]
    assert base.BaseGlyph == "uniE001"
    layers = colr.LayerList.Paint[
        base.Paint.FirstLayerIndex : base.Paint.FirstLayerIndex + base.Paint.NumLayers
    ]
    assert len(layers) == 2
    # check v1 palette indices were remapped
    assert layers[0].Paint.Paint.ColorLine.ColorStop[0].PaletteIndex == 0
    assert layers[0].Paint.Paint.ColorLine.ColorStop[1].PaletteIndex == 1
    assert layers[1].Paint.PaletteIndex == 0

    baseRecV0 = colr.BaseGlyphRecordArray.BaseGlyphRecord[0]
    assert baseRecV0.BaseGlyph == "uniE004"
    layersV0 = colr.LayerRecordArray.LayerRecord
    assert len(layersV0) == 3
    # check v0 palette indices were remapped (except for 0xFFFF)
    assert layersV0[0].PaletteIndex == 0
    assert layersV0[1].PaletteIndex == 0xFFFF
    assert layersV0[2].PaletteIndex == 1

    clipBoxes = colr.ClipList.clips
    assert {"uniE001", "uniE002", "uniE003"} == set(clipBoxes)
    assert clipBoxes["uniE002"] == clipBoxes["uniE003"]

    assert "CPAL" in subset_font
    cpal = subset_font["CPAL"]
    assert [
        tuple(v / 255 for v in (c.red, c.green, c.blue, c.alpha))
        for c in cpal.palettes[0]
    ] == [
        # the first color 'red' was pruned
        (0.0, 1.0, 0.0, 1.0),  # green
        (0.0, 0.0, 1.0, 1.0),  # blue
    ]


def test_subset_COLRv1_and_CPALv1(colrv1_cpalv1_path):
    subset_path = colrv1_cpalv1_path.parent / (colrv1_cpalv1_path.name + ".subset")

    subset.main(
        [
            str(colrv1_cpalv1_path),
            "--glyph-names",
            f"--output-file={subset_path}",
            "--unicodes=E002,E003,E004",
        ]
    )
    subset_font = TTFont(subset_path)

    assert "CPAL" in subset_font
    cpal = subset_font["CPAL"]
    name_table = subset_font["name"]
    assert [
        name_table.getDebugName(name_id) for name_id in cpal.paletteEntryLabels
    ] == [
        # "first color",  # The first color was pruned
        "second color",
        "third color",
    ]
    # check that the "first color" name is dropped from name table
    font = TTFont(colrv1_cpalv1_path)

    first_color_nameID = None
    for n in font["name"].names:
        if n.toUnicode() == "first color":
            first_color_nameID = n.nameID
            break
    assert first_color_nameID is not None
    assert all(n.nameID != first_color_nameID for n in name_table.names)


def test_subset_COLRv1_and_CPALv1_keep_nameID(colrv1_cpalv1_path):
    subset_path = colrv1_cpalv1_path.parent / (colrv1_cpalv1_path.name + ".subset")

    # figure out the name ID of first color so we can keep it
    font = TTFont(colrv1_cpalv1_path)

    first_color_nameID = None
    for n in font["name"].names:
        if n.toUnicode() == "first color":
            first_color_nameID = n.nameID
            break
    assert first_color_nameID is not None

    subset.main(
        [
            str(colrv1_cpalv1_path),
            "--glyph-names",
            f"--output-file={subset_path}",
            "--unicodes=E002,E003,E004",
            f"--name-IDs={first_color_nameID}",
        ]
    )
    subset_font = TTFont(subset_path)

    assert "CPAL" in subset_font
    cpal = subset_font["CPAL"]
    name_table = subset_font["name"]
    assert [
        name_table.getDebugName(name_id) for name_id in cpal.paletteEntryLabels
    ] == [
        # "first color",  # The first color was pruned
        "second color",
        "third color",
    ]

    # Check that the name ID is kept
    assert any(n.nameID == first_color_nameID for n in name_table.names)


def test_subset_COLRv1_and_CPALv1_share_nameID(colrv1_cpalv1_share_nameID_path):
    subset_path = colrv1_cpalv1_share_nameID_path.parent / (
        colrv1_cpalv1_share_nameID_path.name + ".subset"
    )

    subset.main(
        [
            str(colrv1_cpalv1_share_nameID_path),
            "--glyph-names",
            f"--output-file={subset_path}",
            "--unicodes=E002,E003,E004",
        ]
    )
    subset_font = TTFont(subset_path)

    assert "CPAL" in subset_font
    cpal = subset_font["CPAL"]
    name_table = subset_font["name"]
    assert [
        name_table.getDebugName(name_id) for name_id in cpal.paletteEntryLabels
    ] == [
        # "first color",  # The first color was pruned
        "second color",
        "third color",
    ]

    # Check that the name ID 1 is kept
    assert any(n.nameID == 1 for n in name_table.names)


def test_subset_COLRv1_and_CPAL_drop_empty(colrv1_path):
    subset_path = colrv1_path.parent / (colrv1_path.name + ".subset")

    subset.main(
        [
            str(colrv1_path),
            "--glyph-names",
            f"--output-file={subset_path}",
            "--glyphs=glyph00010",
        ]
    )
    subset_font = TTFont(subset_path)

    glyph_set = set(subset_font.getGlyphOrder())

    assert "glyph00010" in glyph_set
    assert "uniE000" not in glyph_set

    assert "COLR" not in subset_font
    assert "CPAL" not in subset_font


def test_subset_COLRv1_downgrade_version(colrv1_path):
    subset_path = colrv1_path.parent / (colrv1_path.name + ".subset")

    subset.main(
        [
            str(colrv1_path),
            "--glyph-names",
            f"--output-file={subset_path}",
            "--unicodes=E004",
        ]
    )
    subset_font = TTFont(subset_path)

    assert set(subset_font.getGlyphOrder()) == {
        ".notdef",
        "uniE004",
        "glyph00016",
        "glyph00017",
        "glyph00018",
    }

    assert "COLR" in subset_font
    assert subset_font["COLR"].version == 0


def test_subset_COLRv1_drop_all_v0_glyphs(colrv1_path):
    subset_path = colrv1_path.parent / (colrv1_path.name + ".subset")

    subset.main(
        [
            str(colrv1_path),
            "--glyph-names",
            f"--output-file={subset_path}",
            "--unicodes=E003",
        ]
    )
    subset_font = TTFont(subset_path)

    assert set(subset_font.getGlyphOrder()) == {
        ".notdef",
        "uniE001",
        "uniE003",
        "glyph00012",
        "glyph00013",
    }

    assert "COLR" in subset_font
    colr = subset_font["COLR"]
    assert colr.version == 1
    assert colr.table.BaseGlyphRecordCount == 0
    assert colr.table.BaseGlyphRecordArray is None
    assert colr.table.LayerRecordArray is None
    assert colr.table.LayerRecordCount == 0


def test_subset_COLRv1_no_ClipList(colrv1_path):
    font = TTFont(colrv1_path)
    font["COLR"].table.ClipList = None  # empty ClipList
    font.save(colrv1_path)

    subset_path = colrv1_path.parent / (colrv1_path.name + ".subset")
    subset.main(
        [
            str(colrv1_path),
            f"--output-file={subset_path}",
            "--unicodes=*",
        ]
    )
    subset_font = TTFont(subset_path)
    assert subset_font["COLR"].table.ClipList is None


def test_subset_keep_size_drop_empty_stylistic_set():
    fb = FontBuilder(unitsPerEm=1000, isTTF=True)
    glyph_order = [".notdef", "a", "b", "b.ss01"]
    fb.setupGlyphOrder(glyph_order)
    fb.setupGlyf({g: TTGlyphPen(None).glyph() for g in glyph_order})
    fb.setupCharacterMap({ord("a"): "a", ord("b"): "b"})
    fb.setupHorizontalMetrics({g: (500, 0) for g in glyph_order})
    fb.setupHorizontalHeader()
    fb.setupOS2()
    fb.setupPost()
    fb.setupNameTable({"familyName": "TestKeepSizeFeature", "styleName": "Regular"})
    fb.addOpenTypeFeatures(
        """
        feature size {
          parameters 10.0 0;
        } size;
        feature ss01 {
          featureNames {
            name "Alternate b";
          };
          sub b by b.ss01;
        } ss01;
    """
    )

    buf = io.BytesIO()
    fb.save(buf)
    buf.seek(0)

    font = TTFont(buf)

    gpos_features = font["GPOS"].table.FeatureList.FeatureRecord
    assert gpos_features[0].FeatureTag == "size"
    assert isinstance(gpos_features[0].Feature.FeatureParams, ot.FeatureParamsSize)
    assert gpos_features[0].Feature.LookupCount == 0
    gsub_features = font["GSUB"].table.FeatureList.FeatureRecord
    assert gsub_features[0].FeatureTag == "ss01"
    assert isinstance(
        gsub_features[0].Feature.FeatureParams, ot.FeatureParamsStylisticSet
    )

    options = subset.Options(layout_features=["*"])
    subsetter = subset.Subsetter(options)
    subsetter.populate(unicodes=[ord("a")])
    subsetter.subset(font)

    # empty size feature was kept
    gpos_features = font["GPOS"].table.FeatureList.FeatureRecord
    assert gpos_features[0].FeatureTag == "size"
    assert isinstance(gpos_features[0].Feature.FeatureParams, ot.FeatureParamsSize)
    assert gpos_features[0].Feature.LookupCount == 0
    # empty ss01 feature was dropped
    assert font["GSUB"].table.FeatureList.FeatureCount == 0


@pytest.mark.skipif(etree is not None, reason="lxml is installed")
def test_subset_svg_missing_lxml(ttf_path):
    # add dummy SVG table and confirm we raise ImportError upon trying to subset it
    font = TTFont(ttf_path)
    font["SVG "] = newTable("SVG ")
    font["SVG "].docList = [('<svg><g id="glyph1"/></svg>', 1, 1)]
    font.save(ttf_path)

    with pytest.raises(ImportError):
        subset.main([str(ttf_path), "--gids=0,1"])


def test_subset_COLR_glyph_closure(tmp_path):
    # https://github.com/fonttools/fonttools/issues/2461
    font = TTFont()
    ttx = pathlib.Path(__file__).parent / "data" / "BungeeColor-Regular.ttx"
    font.importXML(ttx)

    color_layers = font["COLR"].ColorLayers
    assert ".notdef" in color_layers
    assert "Agrave" in color_layers
    assert "grave" in color_layers

    font_path = tmp_path / "BungeeColor-Regular.ttf"
    subset_path = font_path.with_suffix(".subset.ttf)")
    font.save(font_path)

    subset.main(
        [
            str(font_path),
            "--glyph-names",
            f"--output-file={subset_path}",
            "--glyphs=Agrave",
        ]
    )
    subset_font = TTFont(subset_path)

    glyph_order = subset_font.getGlyphOrder()

    assert glyph_order == [
        ".notdef",  # '.notdef' is always included automatically
        "A",
        "grave",
        "Agrave",
        ".notdef.alt001",
        ".notdef.alt002",
        "A.alt002",
        "Agrave.alt001",
        "Agrave.alt002",
        "grave.alt002",
    ]

    color_layers = subset_font["COLR"].ColorLayers
    assert ".notdef" in color_layers
    assert "Agrave" in color_layers
    # Agrave 'glyf' uses grave. It should be retained in 'glyf' but NOT in
    # COLR when we subset down to Agrave.
    assert "grave" not in color_layers


def test_subset_recalc_xAvgCharWidth(ttf_path):
    # Note that the font in in the *ttLib*/data/TestTTF-Regular.ttx file,
    # not this subset/data folder.
    font = TTFont(ttf_path)
    xAvgCharWidth_before = font["OS/2"].xAvgCharWidth

    subset_path = ttf_path.with_suffix(".subset.ttf")
    subset.main(
        [
            str(ttf_path),
            f"--output-file={subset_path}",
            # Keep only the ellipsis, which is very wide, that ought to bump up the average
            "--glyphs=ellipsis",
            "--recalc-average-width",
            "--no-prune-unicode-ranges",
        ]
    )
    subset_font = TTFont(subset_path)
    xAvgCharWidth_after = subset_font["OS/2"].xAvgCharWidth

    # Check that the value gets updated
    assert xAvgCharWidth_after != xAvgCharWidth_before

    # Check that the value gets updated to the actual new value
    subset_font["OS/2"].recalcAvgCharWidth(subset_font)
    assert xAvgCharWidth_after == subset_font["OS/2"].xAvgCharWidth


def test_subset_prune_gdef_markglyphsetsdef():
    # GDEF_MarkGlyphSetsDef
    fb = FontBuilder(unitsPerEm=1000, isTTF=True)
    glyph_order = [
        ".notdef",
        "A",
        "Aacute",
        "Acircumflex",
        "Adieresis",
        "a",
        "aacute",
        "acircumflex",
        "adieresis",
        "dieresiscomb",
        "acutecomb",
        "circumflexcomb",
    ]
    fb.setupGlyphOrder(glyph_order)
    fb.setupGlyf({g: TTGlyphPen(None).glyph() for g in glyph_order})
    fb.setupHorizontalMetrics({g: (500, 0) for g in glyph_order})
    fb.setupHorizontalHeader()
    fb.setupPost()
    fb.setupNameTable(
        {"familyName": "TestGDEFMarkGlyphSetsDef", "styleName": "Regular"}
    )
    fb.addOpenTypeFeatures(
        """
        feature ccmp {
            lookup ccmp_1 {
                lookupflag UseMarkFilteringSet [acutecomb];
                sub a acutecomb by aacute;
                sub A acutecomb by Aacute;
            } ccmp_1;
            lookup ccmp_2 {
                lookupflag UseMarkFilteringSet [circumflexcomb];
                sub a circumflexcomb by acircumflex;
                sub A circumflexcomb by Acircumflex;
            } ccmp_2;
            lookup ccmp_3 {
                lookupflag UseMarkFilteringSet [dieresiscomb];
                sub a dieresiscomb by adieresis;
                sub A dieresiscomb by Adieresis;
                sub A acutecomb by Aacute;
            } ccmp_3;
        } ccmp;
    """
    )

    buf = io.BytesIO()
    fb.save(buf)
    buf.seek(0)

    font = TTFont(buf)

    features = font["GSUB"].table.FeatureList.FeatureRecord
    assert features[0].FeatureTag == "ccmp"
    lookups = font["GSUB"].table.LookupList.Lookup
    assert lookups[0].LookupFlag == 16
    assert lookups[0].MarkFilteringSet == 0
    assert lookups[1].LookupFlag == 16
    assert lookups[1].MarkFilteringSet == 1
    assert lookups[2].LookupFlag == 16
    assert lookups[2].MarkFilteringSet == 2
    marksets = font["GDEF"].table.MarkGlyphSetsDef.Coverage
    assert marksets[0].glyphs == ["acutecomb"]
    assert marksets[1].glyphs == ["circumflexcomb"]
    assert marksets[2].glyphs == ["dieresiscomb"]

    options = subset.Options(layout_features=["*"])
    subsetter = subset.Subsetter(options)
    subsetter.populate(glyphs=["A", "a", "acutecomb", "dieresiscomb"])
    subsetter.subset(font)

    features = font["GSUB"].table.FeatureList.FeatureRecord
    assert features[0].FeatureTag == "ccmp"
    lookups = font["GSUB"].table.LookupList.Lookup
    assert lookups[0].LookupFlag == 16
    assert lookups[0].MarkFilteringSet == 0
    assert lookups[1].LookupFlag == 16
    assert lookups[1].MarkFilteringSet == 1
    marksets = font["GDEF"].table.MarkGlyphSetsDef.Coverage
    assert marksets[0].glyphs == ["acutecomb"]
    assert marksets[1].glyphs == ["dieresiscomb"]

    buf = io.BytesIO()
    fb.save(buf)
    buf.seek(0)

    font = TTFont(buf)

    options = subset.Options(layout_features=["*"], layout_closure=False)
    subsetter = subset.Subsetter(options)
    subsetter.populate(glyphs=["A", "acutecomb", "Aacute"])
    subsetter.subset(font)

    features = font["GSUB"].table.FeatureList.FeatureRecord
    assert features[0].FeatureTag == "ccmp"
    lookups = font["GSUB"].table.LookupList.Lookup
    assert lookups[0].LookupFlag == 16
    assert lookups[0].MarkFilteringSet == 0
    assert lookups[1].LookupFlag == 0
    assert lookups[1].MarkFilteringSet == None
    marksets = font["GDEF"].table.MarkGlyphSetsDef.Coverage
    assert marksets[0].glyphs == ["acutecomb"]


def test_prune_user_name_IDs_with_keep_all(ttf_path):
    font = TTFont(ttf_path)

    keepNameIDs = {n.nameID for n in font["name"].names}

    for i in range(10):
        font["name"].addName(f"Test{i}")

    options = subset.Options()
    options.name_IDs = ["*"]
    options.name_legacy = True
    options.name_languages = ["*"]

    subsetter = subset.Subsetter(options)
    subsetter.populate(unicodes=font.getBestCmap().keys())
    subsetter.subset(font)

    nameIDs = {n.nameID for n in font["name"].names}
    assert not any(n > 255 for n in nameIDs)
    assert nameIDs == keepNameIDs


def test_prune_unused_user_name_IDs_with_keep_all(ttf_path):
    font = TTFont(ttf_path)

    keepNameIDs = {n.nameID for n in font["name"].names}

    for i in range(10):
        font["name"].addName(f"Test{i}")

    nameID = font["name"].addName("Test STAT")
    keepNameIDs.add(nameID)

    font["STAT"] = newTable("STAT")
    font["STAT"].table = ot.STAT()
    font["STAT"].table.ElidedFallbackNameID = nameID

    options = subset.Options()
    options.name_IDs = ["*"]
    options.name_legacy = True
    options.name_languages = ["*"]

    subsetter = subset.Subsetter(options)
    subsetter.populate(unicodes=font.getBestCmap().keys())
    subsetter.subset(font)

    nameIDs = {n.nameID for n in font["name"].names}
    assert nameIDs == keepNameIDs


def test_cvXX_feature_params_nameIDs_are_retained():
    # https://github.com/fonttools/fonttools/issues/3616
    font = TTFont()
    ttx = pathlib.Path(__file__).parent / "data" / "Andika-Regular.subset.ttx"
    font.importXML(ttx)

    keepNameIDs = {n.nameID for n in font["name"].names}

    options = subset.Options()
    options.glyph_names = True
    # that's where the FeatureParamsCharacteVariants are stored
    options.layout_features.append("cv43")

    subsetter = subset.Subsetter(options)
    subsetter.populate(glyphs=font.getGlyphOrder())
    subsetter.subset(font)

    # we expect that all nameIDs are retained, including all the nameIDs
    # used by the FeatureParamsCharacterVariants
    nameIDs = {n.nameID for n in font["name"].names}
    assert nameIDs == keepNameIDs


if __name__ == "__main__":
    sys.exit(unittest.main())
